# coding: utf-8

from time import time
from typing import List, Optional, Tuple
import numpy as np

from .meta_inout import InOut
from .._global import OptionalModule

try:
  from psutil import virtual_memory
except (ModuleNotFoundError, ImportError):
  virtual_memory = OptionalModule("psutil")


class FakeInOut(InOut):
  """This class is a demonstration InOut object that does not require any
  hardware to run.

  It can read and/or modify (to a certain extent) the memory usage of the
  computer.
  
  .. versionadded:: 1.5.5
  .. versionchanged:: 2.0.0 renamed from Fake_inout to FakeInOut
  """

  def __init__(self) -> None:
    """Initializes the parent class."""

    self._buf: Optional[list] = None

    super().__init__()

  def open(self) -> None:
    """Creates the buffer allowing to modify the memory usage."""

    self._buf = list()

  def set_cmd(self, *cmd: float) -> None:
    """Modifies the memory usage of the computer.

    If the command is higher than the current, adds big lists to the buffer in
    order to use more memory.
    If the command is lower than the current, empties the buffer.

    Args:
      *cmd: The target memory usage to set, in percent as a :obj:`float`.
    """

    if not isinstance(cmd[0], float) and not isinstance(cmd[0], int):
      raise TypeError("Not the right command type for the FakeInOut !")
    if not 0 <= cmd[0] <= 100:
      raise ValueError("Command should be a percentage of memory usage !")

    if cmd[0] > virtual_memory().percent:
      self._buf.append([0] * 1024*1024)
    elif cmd[0] < virtual_memory().percent:
      try:
        # If there's nothing to delete, abort
        del self._buf[-1]
      except IndexError:
        return

  def get_data(self) -> List[float]:
    """Just returns the timestamp and the current memory usage."""

    return [time(), virtual_memory().percent]

  def start_stream(self) -> None:
    """Defining this method to avoid getting warnings in the logs.
    
    .. versionadded:: 2.0.0
    """

    ...

  def stop_stream(self) -> None:
    """Defining this method to avoid getting warnings in the logs.

    .. versionadded:: 2.0.0
    """

    ...

  def get_stream(self) -> Tuple[np.ndarray, np.ndarray]:
    """This method calls 10 times the :meth:`get_data` method and returns the
    10 values at once in the streamer format.

    It is just a demo for showcasing the use of the streamer mode.
    
    .. versionadded:: 2.0.0
    """

    values = np.array([self.get_data() for _ in range(10)])
    t, ret = np.split(values, 2, axis=1)
    return np.squeeze(t), ret

  def close(self) -> None:
    """Deletes the buffer."""

    del self._buf



# coding: utf-8

from time import time
from typing import Union, Dict
from itertools import cycle, islice
import logging

from .meta_path import Path, ConditionType


class Cyclic(Path):
  """This Path cyclically alternates between two constant values, based on two
  different conditions.

  It can for example be used as a trigger, or used to drive an actuator
  cyclically. It is equivalent to a succession of 
  :class:`~crappy.blocks.generator_path.Constant` Paths.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               condition1: Union[str, ConditionType],
               condition2: Union[str, ConditionType],
               value1: float,
               value2: float,
               cycles: float = 1) -> None:
    """Sets the arguments and initializes the parent class.

    The Path always starts with ``value1``, and then switches to ``value2``.

    Args:
      condition1: The condition for switching to ``value2``. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      condition2: The condition for switching to ``value1``. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      value1: First value to send.
      value2: Second value to send.
      cycles: Number of cycles. Half cycles are accepted. If `0`, loops
        forever.

    Note:
      ::

        [{'type': 'Cyclic', 'value1': 1, 'condition1': 'AIN0>2',
        'value2': 0, 'condition2': 'AIN1<1', 'cycles': 5}]

      is equivalent to
      ::

        [{'type': 'Constant', 'value': 1,'condition': 'AIN0>2'},
        {'type': 'Constant', 'value': 0, 'condition': 'AIN1<1'}] * 5
    
    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionremoved:: 1.5.10 *verbose* argument
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    super().__init__()

    # Creates an interator object with a given length
    if cycles > 0:
      cycles = int(2 * cycles)
      self._values = islice(cycle((value1, value2)), cycles)
      self._conditions = islice(cycle((self.parse_condition(condition1),
                                       self.parse_condition(condition2))),
                                cycles)

    # Creates an endless iterator object
    else:
      self._values = cycle((value1, value2))
      self._conditions = cycle((self.parse_condition(condition1),
                                self.parse_condition(condition2)))

    # The current condition object and value
    self._condition = None
    self._value = None

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns either the first or second value depending on the current state
    of the cycle. Raises :exc:`StopIteration` when the cycles are exhausted.

    Also manages the switch between the values and conditions 1 and 2.
    """

    # During the first loop, getting the first condition and value
    if self._value is None and self._condition is None:
      try:
        self._value = next(self._values)
        self._condition = next(self._conditions)
        self.log(logging.DEBUG, f"Got value {self._value} and condition "
                                f"{self._condition}")
      except StopIteration:
        self.log(logging.DEBUG, "No value or condition, switching to next "
                                "path")
        raise

    # During other loops, getting the next condition and value if the current
    # condition is met
    if self._condition(data):
      self.log(logging.DEBUG, "Ended phase of the cycle, switching to the "
                              "next phase")
      self.t0 = time()
      try:
        self._value = next(self._values)
        self._condition = next(self._conditions)
        self.log(logging.DEBUG, f"Got value {self._value} and condition "
                                f"{self._condition}")
      except StopIteration:
        self.log(logging.DEBUG, "Stop condition met, switching to next path")
        raise

    # Finally, returning the current value
    return self._value



# coding: utf-8

from .meta_block import Block


def recv_all(_):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method recv_all was deprecated in version "
                            "2.0.0, please use recv_all_data instead !")


def poll(_):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method poll was deprecated in version "
                            "2.0.0, please use data_available instead !")


def recv_all_last(_):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method recv_all_last was deprecated in "
                            "version 2.0.0, please use recv_last_data "
                            "instead !")


def get_last(_, *__, **___):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method get_last was deprecated in version "
                            "2.0.0, please use recv_last_data instead !")


def get_all_last(_, *__, **___):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method get_all_last was deprecated in "
                            "version 2.0.0, please use recv_all_data "
                            "instead !")


def recv_all_delay(_, *__, **___):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method recv_all_delay was deprecated in "
                            "version 2.0.0, please use recv_all_data_raw "
                            "instead !")


def drop(_, *__, **___):
  """Empty function for signaling a deprecated method of the Block object."""

  raise NotImplementedError("The method drop was deprecated in version "
                            "2.0.0 !")


setattr(Block, recv_all.__name__, recv_all)
setattr(Block, poll.__name__, poll)
setattr(Block, recv_all_last.__name__, recv_all_last)
setattr(Block, get_last.__name__, get_last)
setattr(Block, get_all_last.__name__, get_all_last)
setattr(Block, recv_all_delay.__name__, recv_all_delay)
setattr(Block, drop.__name__, drop)


class AutoDrive(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to AutoDriveVideoExtenso in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Client_server(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to ClientServer in version 2.0.0 ! "
                              f"Check the documentation for more information.")
  
  
class Displayer(Block):
  """Empty class for signaling an object of version 1.5 that wa removed in 
  version 2.0.O."""

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was removed "
                              f"in version 2.0.0 ! It is now contained "
                              f"directly in the Camera Block. "
                              f"Check the documentation for more information.")


class DISVE(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to DICVE in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Drawing(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to Canvas in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Fake_machine(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to FakeMachine in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class GUI(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to Button in version 2.0.0 ! "
                              f"Check the documentation for more information.")
  
  
class Hdf_recorder(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to HDFRecorder in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Mean_block(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to MeanBlock in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Multiplex(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to Multiplexer in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Reader(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to LinkReader in version 2.0.0 ! "
                              f"Check the documentation for more information.")


class Video_extenso(Block):
  """Empty class for signaling an object of version 1.5 whose name changed in
  version 2.0 and is now deprecated.

  The new name of the correct object to use is given.
  """

  def __init__(self, *_, **__) -> None:
    """Simply raises the exception when instantiating the object."""

    super().__init__()

    raise NotImplementedError(f"The {type(self).__name__} Block was renamed "
                              f"to VideoExtenso in version 2.0.0 ! "
                              f"Check the documentation for more information.")



# coding: utf-8

from time import time
from typing import Union, Optional, List
import logging

from .meta_inout import InOut
from .._global import OptionalModule

try:
  import adafruit_mprls
except (ImportError, ModuleNotFoundError):
  adafruit_mprls = OptionalModule('adafruit_mprls')

try:
  import board
except (ImportError, ModuleNotFoundError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  from digitalio import DigitalInOut
except (ImportError, ModuleNotFoundError):
  DigitalInOut = OptionalModule('digitalio',
                                'Blinka is necessary to access the GPIOs')

try:
  from smbus2 import SMBus, i2c_msg
except (ImportError, ModuleNotFoundError):
  SMBus = i2c_msg = OptionalModule('smbus2')

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

mprls_status_bits = {'busy': 0x20,
                     'memory error': 0x04,
                     'math saturation': 0x01}
mprls_backends = ['Pi4', 'blinka']


class MPRLS(InOut):
  """This class can read values from an MPRLS pressure sensor.

  It communicates over I2C with the sensor.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Mprls to MPRLS
  """

  def __init__(self,
               backend: str,
               eoc_pin: Optional[Union[str, int]] = None,
               device_address: int = 0x18,
               i2c_port: int = 1) -> None:
    """Initializes the parent class and opens the I2C bus.

    Args:
      backend: Should be one of :
        ::

          'Pi4', 'blinka'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing
        :mod:`Adafruit-Blinka` and :mod:`adafruit-circuitpython-mprls`, but
        these modules are compatible with and maintained on a wide
        variety of boards.
      eoc_pin: Optionally, reads the end of conversion signal from the polarity
        of a GPIO rather than from an I2C register. Speeds up the reading and
        decreases the traffic on the bus, but requires one extra wire. With the
        backend `'Pi4'`, give the index of the GPIO in BCM convention. With the
        backend `'blinka'`, it should be a string but the syntax varies
        according to the board. Refer to blinka's documentation for more
        information.
      device_address: The I2C address of the MPRLS. The address of the devices
        sold by Adafruit is `0x18`, but other suppliers may sell it with
        another address.
      i2c_port: The I2C port over which the MPRLS should communicate. On most
        Raspberry Pi models the default I2C port is `1`.
    
    .. versionadded:: 1.5.8
       *backend*, *eoc_pin*, *device_address*, *i2c_port* and *ft232h_ser_num*
       arguments
    .. versionremoved:: 2.0.0 *ft232h_ser_num* argument
    """

    self._bus = None
    self._eoc_pin = None
    self._mpr = None
    self._i2c_msg = None

    if not isinstance(backend, str) or backend not in mprls_backends:
      raise ValueError("backend should be in {}".format(mprls_backends))
    self._backend = backend

    super().__init__()

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer.")
    self._address = device_address

    if not isinstance(i2c_port, int):
      raise TypeError("i2c_port should be an integer.")
    self._i2c_port = i2c_port

    if eoc_pin is not None:
      if backend == 'blinka' and not isinstance(eoc_pin, str):
        raise TypeError('eoc_pin should be a string when using the blinka '
                        'backend !')
      elif backend == 'ft232h' and not isinstance(eoc_pin, str):
        raise TypeError('eoc_pin should be a string when using the ft232h '
                        'backend !')
      elif backend == 'Pi4' and not isinstance(eoc_pin, int):
        raise TypeError('eoc_pin should be an int when using the Pi4 '
                        'backend !')
    self._eoc_pin = eoc_pin

  def open(self) -> None:
    """Opens the I2C bus."""

    if self._backend == 'Pi4':
      self.log(logging.INFO, f"Opening the I2C connection to the MPRLS on "
                             f"port {self._i2c_port}")
      self._bus = SMBus(self._i2c_port)
      if self._eoc_pin is not None:
        self.log(logging.INFO, "Setting up the GPIOs")
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._eoc_pin, GPIO.IN)

    elif self._backend == 'blinka':
      if self._eoc_pin is not None:
        eoc = DigitalInOut(getattr(board, self._eoc_pin))
      else:
        eoc = None
      self.log(logging.INFO, "Connecting the the MPRLS with backend blinka")
      self._mpr = adafruit_mprls.MPRLS(board.I2C(),
                                       psi_min=0,
                                       psi_max=25,
                                       eoc_pin=eoc)

    self._i2c_msg = i2c_msg

  def get_data(self) -> List[float]:
    """Reads the pressure value.

    Returns:
      The timestamp and the pressure value in hPa.
    """

    if self._backend == 'blinka':
      pres = self._mpr.pressure

    else:
      # Starting conversion
      self.log(logging.DEBUG, f"Writing {0xAA, 0x00, 0x00} to the address "
                              f"{self._address}")
      self._bus.i2c_rdwr(self._i2c_msg.write(self._address,
                                             [0xAA, 0x00, 0x00]))
      # Waiting for conversion to complete
      t0 = time()
      while not self._data_available():
        if time() - t0 > 0.1:
          raise TimeoutError('Waited too long for data to be ready')

      # Reading conversion result
      read = self._i2c_msg.read(self._address, 4)
      self._bus.i2c_rdwr(read)
      out = list(read)
      self.log(logging.DEBUG, f"Read {out} from address {self._address}")
      # Checking if anu error occurred
      if out[0] & mprls_status_bits['memory error']:
        raise RuntimeError("A memory error occurred on the MPRLS")
      elif out[0] & mprls_status_bits['math saturation']:
        raise RuntimeError("A math saturation error occurred on the MPRLS")
      # Extracting conversion result as an integer
      ret = (out[1] << 16) | (out[2] << 8) | out[3]
      # Converting to hPa
      pres = 68.947572932 * (ret - 0x19999A) * 25 / (0xE66666 - 0x19999A)

    return [time(), pres]

  def close(self) -> None:
    """Closes the I2C bus."""

    if self._backend == 'Pi4' and self._bus is not None:
      self.log(logging.INFO, "Closing the I2C connection to the MPRLS")
      self._bus.close()

    if self._backend == 'Pi4' and self._eoc_pin is not None:
      self.log(logging.INFO, "Cleaning up the GPIOs")
      GPIO.cleanup()

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._eoc_pin is None:
      wait = self._i2c_msg.read(self._address, 1)
      self._bus.i2c_rdwr(wait)
      return not list(wait)[0] & mprls_status_bits['busy']
    else:
      return bool(GPIO.input(self._eoc_pin))



# coding: utf-8

from dataclasses import dataclass


@dataclass
class Zoom:
  """This class stores the upper and lower limits of the image to display in
  the :class:`~crappy.tool.camera_config.CameraConfig` window.

  It also allows updating them when the user changes the zoom ratio or drags
  the image with the mouse.

  .. versionadded:: 2.0.0
  """

  x_low: float = 0.
  x_high: float = 1.
  y_low: float = 0.
  y_high: float = 1.

  def reset(self) -> None:
    """Resets the zoom level to default (no zoom)."""

    self.x_low, self.x_high, self.y_low, self.y_high = 0, 1, 0, 1

  def update_zoom(self, x: float, y: float, ratio: float) -> None:
    """Updates the upper and lower limits of the image when the user scrolls
    with the mousewheel.

    The update is based on the zoom ratio and the position of the mouse on the
    screen.

    Args:
      x: The `x` position of the mouse on the image, as a ratio between `0`
        and `1`.
      y: The `y` position of the mouse on the image, as a ratio between `0`
        and `1`.
      ratio: The zoom ratio to apply. If it is greater than `1` we zoom in,
        otherwise we zoom out.
    """

    prev_x_low, prev_x_high = self.x_low, self.x_high
    prev_y_low, prev_y_high = self.y_low, self.y_high

    # Updating the lower x limit
    self.x_low = max(self.x_low + x * (1 - 1 / ratio), 0.)
    # Updating the upper x limit, making sure it's not out of the image
    if self.x_low + 1 / ratio * (prev_x_high - prev_x_low) > 1.:
      self.x_high = 1.
      self.x_low = 1 - 1 / ratio * (prev_x_high - prev_x_low)
    else:
      self.x_high = self.x_low + 1 / ratio * (prev_x_high - prev_x_low)

    # Updating the lower y limit
    self.y_low = max(self.y_low + y * (1 - 1 / ratio), 0.)
    # Updating the upper y limit, making sure it's not out of the image
    if self.y_low + 1 / ratio * (prev_y_high - prev_y_low) > 1.:
      self.y_high = 1.
      self.y_low = 1 - 1 / ratio * (prev_y_high - prev_y_low)
    else:
      self.y_high = self.y_low + 1 / ratio * (prev_y_high - prev_y_low)

  def update_move(self, delta_x: float, delta_y: float) -> None:
    """Updates the upper and lower limits of the image when the user moves the
    image with a left button click.

    Args:
      delta_x: The `x` displacement to apply to the image, as a ratio of the
        total image width.
      delta_y: The `y` displacement to apply to the image, as a ratio of the
        total image height.
    """

    prev_x_low, prev_x_high = self.x_low, self.x_high
    prev_y_low, prev_y_high = self.y_low, self.y_high

    # Updating the x position
    if delta_x <= 0:
      self.x_low = max(0., prev_x_low + delta_x)
      self.x_high = self.x_low + prev_x_high - prev_x_low
    else:
      self.x_high = min(1., prev_x_high + delta_x)
      self.x_low = self.x_high - prev_x_high + prev_x_low

    # Updating the y position
    if delta_y <= 0:
      self.y_low = max(0., prev_y_low + delta_y)
      self.y_high = self.y_low + prev_y_high - prev_y_low
    else:
      self.y_high = min(1., prev_y_high + delta_y)
      self.y_low = self.y_high - prev_y_high + prev_y_low



# coding: utf-8

from time import time
import logging
from typing import Optional

from .meta_block import Block
from .._global import OptionalModule

try:
  import tkinter as tk
except (ModuleNotFoundError, ImportError):
  tk = OptionalModule("tkinter")


class Button(Block):
  """This Block allows the user to send a signal to downstream Blocks upon
  clicking on a button in a Graphical User Interface.

  It sends an integer value, that starts from `0` and is incremented every time
  the user clicks on the button. This Block relies on a :obj:`~tkinter.Tk`
  window for the graphical interface.

  This Block is mostly useful for incorporating user feedback in a script, i.e.
  triggering actions based on an experimenter's decision. It can be handy for
  taking pictures at precise moments, or when an action should only begin after
  the experimenter has completed a task, for example.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from GUI to Button
  """

  def __init__(self,
               send_0: bool = False,
               label: str = 'step',
               time_label: str = 't(s)',
               freq: Optional[float] = 50,
               spam: bool = False,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      send_0: If :obj:`True`, the value `0` will be sent automatically when
        starting the Block. Otherwise, `1` will be sent at the first click.
        Only relevant when ``spam`` is :obj:`False`.
      label: The label carrying the information on the number of clicks,
        default is ``'step'``.
      time_label: The label carrying the time information, default is
        ``'t(s)'``.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      spam: If :obj:`True`, sends the current step value at each loop,
        otherwise only sends it at each click.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded:: 1.5.10 *send_0* and *time_label* arguments
    .. versionadded:: 2.0.0 *display_freq* and *debug* arguments
    """

    self._root: Optional[tk.Tk] = None

    super().__init__()
    self.freq = freq
    self.labels = [time_label, label]
    self.display_freq = display_freq
    self.debug = debug

    self._spam = spam
    self._send_0 = send_0

    self._step = None
    self._text = None
    self._label = None
    self._button = None

  def prepare(self) -> None:
    """Creates the graphical interface and sets its layout and callbacks."""

    self.log(logging.INFO, "Creating the GUI")

    self._root = tk.Tk()
    self._root.title("Button block")
    self._root.resizable(False, False)

    self._step = tk.IntVar()
    self._step.trace_add('write', self._update_text)
    self._text = tk.StringVar(value=f'step: {self._step.get()}')

    self._label = tk.Label(self._root, textvariable=self._text)
    self._label.pack(padx=7, pady=7)

    self._button = tk.Button(self._root,
                             text='Next step',
                             command=self._next_step)
    self._button.pack(padx=25, pady=7)

    self._root.update()

  def begin(self) -> None:
    """Sends the value of the first step (`0`) if required."""

    if self._send_0:
      self.send([time() - self.t0, self._step.get()])

  def loop(self) -> None:
    """Updates the interface, and sends the current step value if ``spam`` is
    :obj:`True`. """

    try:
      self._root.update()
      self.log(logging.DEBUG, "GUI updated")
    except tk.TclError:
      return

    if self._spam:
      self.send([time() - self.t0, self._step.get()])

  def finish(self) -> None:
    """Closes the interface window."""

    self.log(logging.INFO, "closing the GUI")
    try:
      if self._root is not None:
        self._root.destroy()
    except tk.TclError:
      pass

  def _update_text(self, _, __, ___) -> None:
    """Simply updates the displayed text."""

    self._text.set(f'step: {self._step.get()}')

  def _next_step(self) -> None:
    """Increments the step counter and sends the corresponding signal."""

    self.log(logging.DEBUG, "Next step on the GUI")
    self._step.set(self._step.get() + 1)
    self.send([time() - self.t0, self._step.get()])



# coding: utf-8

from time import time
from typing import Tuple, Optional
from numpy import ndarray
import logging
from .meta_camera import Camera
from .._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class CameraOpencv(Camera):
  """A class for reading images from any camera able to interface with OpenCV.

  The number of the video device to read images from can be specified.

  This camera class is less performant than the
  :class:`~crappy.camera.CameraGstreamer` one that relies on GStreamer, but the
  installation of OpenCV is way easier than the one of GStreamer, especially on
  Windows !

  Warning:
    There are two classes for CameraOpencv, one for Linux based on
    `v4l-utils`, and another one for Linux (without `v4l-utils`) and other OS.
    Depending on the installation of `v4l-utils` and the OS, the correct class
    will be automatically imported. The version using `v4l-utils` allows tuning
    more parameters than the basic version.

  .. versionadded:: 1.5.9
  .. versionchanged:: 2.0.0 renamed from Camera_opencv to CameraOpencv
  """

  def __init__(self) -> None:
    """Sets variables and adds the channels setting."""

    super().__init__()

    self._cap = None

    self.add_choice_setting(name="channels",
                            choices=('1', '3'),
                            default='1')

  def open(self, device_num: Optional[int] = 0, **kwargs) -> None:
    """Opens the video stream and sets any user-specified settings.

    Args:
      device_num: The index of the device to open, as an :obj:`int`.
      **kwargs: Any additional setting to set before opening the configuration
        window.
    """

    # Opening the videocapture device
    self.log(logging.INFO, "Opening the image stream from the camera")
    self._cap = cv2.VideoCapture(device_num)

    min_bright, max_bright = self._get_min_max(cv2.CAP_PROP_BRIGHTNESS)
    if min_bright != max_bright:
      self.add_scale_setting(name='brightness',
                             lowest=min_bright,
                             highest=max_bright,
                             getter=self._get_brightness,
                             setter=self._set_brightness)
    else:
      self.log(logging.INFO, "Could not add Brightness as scale setting,"
                             "the detected min and max values are identical")

    min_cont, max_cont = self._get_min_max(cv2.CAP_PROP_CONTRAST)
    if min_cont != max_cont:
      self.add_scale_setting(name='contrast',
                             lowest=min_cont,
                             highest=max_cont,
                             getter=self._get_contrast,
                             setter=self._set_contrast)
    else:
      self.log(logging.INFO, "Could not add Contrast as scale setting,"
                             "the detected min and max values are identical")

    min_hue, max_hue = self._get_min_max(cv2.CAP_PROP_HUE)
    if min_hue != max_hue:
      self.add_scale_setting(name='hue',
                             lowest=min_hue,
                             highest=max_hue,
                             getter=self._get_hue,
                             setter=self._set_hue)
    else:
      self.log(logging.INFO, "Could not add HUE as scale setting,"
                             "the detected min and max values are identical")

    min_sat, max_sat = self._get_min_max(cv2.CAP_PROP_SATURATION)
    if min_sat != max_sat:
      self.add_scale_setting(name='saturation',
                             lowest=min_sat,
                             highest=max_sat,
                             getter=self._get_saturation,
                             setter=self._set_saturation)
    else:
      self.log(logging.INFO, "Could not add Saturation as scale setting,"
                             "the detected min and max values are identical")

    # Setting the kwargs if any
    self.set_all(**kwargs)

  def get_image(self) -> Tuple[float, ndarray]:
    """Grabs a frame from the videocapture object and returns it along with a
    timestamp."""

    # Grabbing the frame and the timestamp
    t = time()
    ret, frame = self._cap.read()

    # Checking the integrity of the frame
    if not ret:
      raise IOError("Error reading the camera")

    # Returning the image in the right format, and its timestamp
    if self.channels == '1':
      return t, cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    else:
      return t, frame

  def close(self) -> None:
    """Releases the videocapture object."""

    if self._cap is not None:
      self.log(logging.INFO, "Closing the image stream from the camera")
      self._cap.release()

  def _get_min_max(self, prop_id: int) -> Tuple[int, int]:
    """Gets the min and max values of a parameter."""

    self._cap.set(prop_id, 99999)
    max_ = int(self._cap.get(prop_id))
    self._cap.set(prop_id, -99999)
    min_ = int(self._cap.get(prop_id))
    if min_ == max_:
      self._cap.set(prop_id, 0)
      min_ = int(self._cap.get(prop_id))
    return min_, max_

  def _get_brightness(self) -> int:
    """Gets the image brightness."""

    return int(self._cap.get(cv2.CAP_PROP_BRIGHTNESS))

  def _get_contrast(self) -> int:
    """Gets the image contrast."""

    return int(self._cap.get(cv2.CAP_PROP_CONTRAST))

  def _get_hue(self) -> int:
    """Gets the image hue."""

    return int(self._cap.get(cv2.CAP_PROP_HUE))

  def _get_saturation(self) -> int:
    """Gets the image saturation."""

    return int(self._cap.get(cv2.CAP_PROP_SATURATION))

  def _set_brightness(self, brightness: int) -> None:
    """Sets the image brightness."""

    self._cap.set(cv2.CAP_PROP_BRIGHTNESS, brightness)

  def _set_contrast(self, contrast: int) -> None:
    """Sets the image contrast."""

    self._cap.set(cv2.CAP_PROP_CONTRAST, contrast)

  def _set_hue(self, hue: int) -> None:
    """Sets the image hue."""

    self._cap.set(cv2.CAP_PROP_HUE, hue)

  def _set_saturation(self, saturation: int) -> None:
    """Sets the image saturation."""

    self._cap.set(cv2.CAP_PROP_SATURATION, saturation)



# coding: utf-8

from typing import Callable, Union, Optional, Tuple
from pathlib import Path
import numpy as np
from time import time, sleep, strftime, gmtime
from types import MethodType
from multiprocessing import Array, Manager, Event, RLock, Pipe, Barrier
from multiprocessing.sharedctypes import SynchronizedArray
from multiprocessing import managers, synchronize, connection
from threading import BrokenBarrierError
import logging

from .meta_block import Block
from .camera_processes import Displayer, ImageSaver, CameraProcess
from ..camera import camera_dict, Camera as BaseCam, deprecated_cameras
from ..tool.camera_config import CameraConfig
from .._global import CameraPrepareError, CameraRuntimeError, CameraConfigError


class Camera(Block):
  """This Block can drive a :class:`~crappy.camera.Camera` object. It can 
  acquire images, display them and record them. It can only drive one Camera at
  once.
  
  It takes no input :class:`~crappy.links.Link` in a majority of situations,
  and never has output Links. Most of the time, this Block is used for 
  recording to the desired location the images it acquires. Optionally, the 
  images can also be displayed in a dedicated window. Both of these features 
  are however optional, and it is possible to acquire images and not do 
  anything with them. Several options are available for tuning the record and 
  the display.
  
  Before a test starts, this Block can also display a 
  :class:`~crappy.tool.camera_config.CameraConfig` window in which the user can
  visualize the acquired images, and interactively tune all the 
  :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting` available
  for the instantiated :class:`~crappy.camera.Camera`. 
  
  Internally, this Block is only in charge of the image acquisition, and the 
  other tasks are parallelized and delegated to 
  :class:`~crappy.blocks.camera_processes.CameraProcess` objects. The display 
  is handled by the :class:`~crappy.blocks.camera_processes.Displayer`, and
  the recording by the :class:`~crappy.blocks.camera_processes.ImageSaver`.
  This Block manages the instantiation, the synchronisation and the
  termination of all the CameraProcess it controls.
  
  .. versionadded:: 1.4.0
  """

  cam_count = dict()

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               software_trig_label: Optional[str] = None,
               display_freq: bool = False,
               debug: Optional[bool] = False,
               freq: Optional[float] = 200,
               save_images: bool = False,
               img_extension: str = "tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               img_shape: Optional[Tuple[int, int]] = None,
               img_dtype: Optional[str] = None,
               **kwargs) -> None:
    """Sets the arguments and initializes the parent class.
    
    Args:
      camera: The name of the :class:`~crappy.camera.Camera` object to use for
        acquiring the images. Arguments can be passed to this Camera as 
        ``kwargs`` of this Block. This argument is ignored if the 
        ``image_generator`` argument is provided.
      transform: A callable taking an image as an argument, and returning a
        transformed image as an output. Allows applying a post-processing
        operation to the acquired images. This is done right after the
        acquisition, so the original image is permanently lost and only the
        transformed image is displayed and/or saved and/or further processed.
        The transform operation is not parallelized, so it might negatively
        affect the acquisition framerate if it is too heavy.
      config: If :obj:`True`, a 
        :class:`~crappy.tool.camera_config.CameraConfig` window is displayed
        before the test starts. There, the user can interactively adjust the 
        different 
        :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting` 
        available for the selected :class:`~crappy.camera.Camera`, and 
        visualize the acquired images. The test starts when closing the 
        configuration window. If not enabled, the ``img_dtype`` and 
        ``img_shape`` arguments must be provided.
      display_images: If :obj:`True`, displays the acquired images in a
        dedicated window, using the backend given in ``displayer_backend`` and
        at the frequency specified in ``displayer_framerate``. This option
        should be considered as a debug or basic follow-up feature, it is not
        intended to be very fast nor to display high-quality images. The
        maximum resolution of the displayed images in `640x480`, the images
        might be downscaled to fit in this format.
      displayer_backend: The backend to use for displaying the images. Can be
        either ``'cv2'`` or ``'mpl'``, to use respectively :mod:`cv2` (OpenCV)
        or :mod:`matplotlib`. ``'cv2'`` usually allows achieving a higher
        display frequency. Ignored if ``display_images`` is :obj:`False`. If
        not given and ``display_images`` is :obj:`True`, ``'cv2'`` is tried
        first and ``'mpl'`` second, and the first available one is used.
      displayer_framerate: The maximum update frequency of the image displayer, 
        as an :obj:`int`. This value usually lies between 5 and 30Hz, the 
        default is 5. The achieved update frequency might be lower than
        requested. Ignored if ``display_images`` is :obj:`False`.
      software_trig_label: The name of a label used as a software trigger for 
        the :class:`~crappy.camera.Camera`. If given, images will only be 
        acquired when receiving data over this label. The received value does
        not matter. This software trigger is not meant to be very precise, it
        is recommended not to rely on it for a trigger frequency greater than
        10Hz, in which case a hardware trigger should be preferred if available
        on the camera.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      save_images: If :obj:`True`, the acquired images are saved to the folder
        specified in ``save_folder``, in the format specified in 
        ``img_extension``, using the backend specified in ``save_backend``, and
        at the frequency specified in ``save_period``. Each image is saved with
        the name : ``<frame_nr>_<timestamp>.<extension>``, and can thus easily
        be identified. Along with the images, a ``metadata.csv`` file records 
        the metadata of all the saved images. This metadata is either the one 
        returned by the :meth:`~crappy.camera.Camera.get_image` method of the
        :class:`~crappy.camera.Camera` object, or the default one generated in
        the :meth:`loop` method of this Block. Depending on the framerate of
        the camera and the performance of the computer, it is not guaranteed 
        that all the acquired images will be recorded.
      img_extension: The file extension for the recorded images, as a
        :obj:`str` and without the dot. Common file extensions include `tiff`,
        `png`, `jpg`, etc. Depending on the used ``save_backend``, some 
        extensions might not be available. It is currently not possible to
        customize the save parameters further than choosing the file extension.
        Ignored if ``save_images`` is :obj:`False`.
      save_folder: Path to the folder where to save the images, either as a 
        :obj:`str` or as a :obj:`pathlib.Path`. Can be an absolute or a 
        relative path, pointing to a folder. If the folder does not exist, it 
        will be created (if the user has permission). If the given folder 
        already contains a ``metadata.csv`` file (and thus likely images from
        Crappy), images are saved to another folder with the same name except
        a suffix is appended. Ignored if ``save_images`` is :obj:`False`. If
        not provided and ``save_images`` is :obj:`True`, the images are saved
        to the folder ``Crappy_images``, created next to the running script.
      save_period: Must be given as an :obj:`int`. Only one out of that number 
        images at most will be saved. Allows to have a known periodicity in 
        case the framerate is too high to record all the images. Or simply to 
        reduce the number of recorded images if saving them all is not needed.
        Ignored if ``save_images`` is :obj:`False`.
      save_backend: If ``save_images`` is :obj:`True`, the backend to use for 
        recording the images. It should be one of:
        ::

          'sitk', 'cv2', 'pil', 'npy'
        
        They correspond to the modules :mod:`SimpleITK`, :mod:`cv2` (OpenCV),
        :mod:`PIL` (Pillow Fork), and :mod:`numpy`. Note that the ``'npy'``
        backend saves the images as raw :obj:`numpy.array`, and thus ignores
        the ``img_extension`` argument. Depending on the machine, some backends
        may be faster or slower. For using each backend, the corresponding 
        Python must of course be installed. If not provided and ``save_images``
        is :obj:`True`, the backends are tried in the same order as given above
        and the first available one is used. ``'npy'`` is always available.
      image_generator: A callable taking two :obj:`float` as arguments and
        returning an image as a :obj:`numpy.array`. **This argument is intended
        for use in the examples of Crappy, to apply an artificial strain on a
        base image. Most users should ignore it.** When given, the ``camera``
        argument is ignored and the images are acquired from the generator. To
        apply a strain on the image, strain values (in `%`) should be sent to 
        the Camera Block over the labels ``'Exx(%)'`` and ``'Eyy(%)'``.
      img_shape: The shape of the images returned by the 
        :class:`~crappy.camera.Camera` object as a :obj:`tuple` of :obj:`int`.
        It should correspond to the value returned by :obj:`numpy.shape`. 
        **This argument is mandatory in case** ``config`` **is** :obj:`False`.
        It is otherwise ignored.
      img_dtype: The `dtype` of the images returned by the
        :class:`~crappy.camera.Camera` object, as a :obj:`str`. It should
        correspond to a valid data type in :mod:`numpy`, e.g. ``'uint8'``.
        **This argument is mandatory in case** ``config`` **is** :obj:`False`.
        It is otherwise ignored.
      **kwargs: Any additional argument will be passed to the 
        :class:`~crappy.camera.Camera` object, and used as a kwarg to its
        :meth:`~crappy.camera.Camera.open` method.
    
    .. versionadded:: 1.5.2 *no_loop* argument
    .. versionadded:: 1.5.10 
       *display_images*, *displayer_backend*, *displayer_framerate*, 
       *software_trig_label*, *freq*, *save_images* and *image_generator*
       arguments
    .. versionremoved::
       1.5.10 *fps_label*, *ext*, *input_label* and *no_loop* arguments
    .. versionadded:: 2.0.0
       *debug*, *img_extension*, *img_shape* and *img_dtype* arguments
    .. versionremoved:: 2.0.0 *img_name* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._save_proc: Optional[ImageSaver] = None
    self._display_proc: Optional[Displayer] = None
    self.process_proc: Optional[CameraProcess] = None
    self._manager: Optional[managers.SyncManager] = None

    self._camera: Optional[BaseCam] = None

    super().__init__()

    self.display_freq = display_freq
    self.freq = freq
    self.niceness = -10
    self.debug = debug

    # Checking for deprecated names
    if camera in deprecated_cameras:
      raise NotImplementedError(
          f"The {camera} Camera was deprecated in version 2.0.0, and renamed "
          f"to {deprecated_cameras[camera]} ! Please update your code "
          f"accordingly and check the documentation for more information")

    # Checking if the requested camera exists in Crappy
    if image_generator is None:
      if camera not in camera_dict:
        possible = ', '.join(sorted(camera_dict.keys()))
        raise ValueError(f"Unknown Camera type : {camera} ! "
                         f"The possible types are : {possible}")
      self._camera_name = camera
    else:
      self._camera_name = 'Image Generator'

    # Counting the number of instantiated cameras for each type
    if self._camera_name not in Camera.cam_count:
      Camera.cam_count[self._camera_name] = 1
    else:
      Camera.cam_count[self._camera_name] += 1

    # Setting the other attributes
    self._trig_label = software_trig_label
    self._config_cam = config
    self._transform = transform
    self._image_generator = image_generator
    self._img_shape = img_shape
    self._img_dtype = img_dtype
    self._camera_kwargs = kwargs

    # The synchronization objects are initialized later
    self._img_array: Optional[SynchronizedArray] = None
    self._img: Optional[np.ndarray] = None
    self._metadata: Optional[managers.DictProxy] = None
    self._cam_barrier: Optional[synchronize.Barrier] = None
    self._stop_event_cam: Optional[synchronize.Event] = None
    self._overlay_conn_in: Optional[connection.Connection] = None
    self._overlay_conn_out: Optional[connection.Connection] = None
    self._save_lock: Optional[synchronize.RLock] = None
    self._disp_lock: Optional[synchronize.RLock] = None
    self._proc_lock: Optional[synchronize.RLock] = None

    self._loop_count = 0
    self._fps_count = 0
    self._last_cam_fps = time()

    # Instantiating the ImageSaver if requested
    self._save_images = save_images
    self._img_extension = img_extension
    self._save_folder = save_folder
    self._save_period = save_period
    self._save_backend = save_backend

    # Instantiating the Displayer window if requested
    self._display_images = display_images
    self._title = f"Displayer {camera} {Camera.cam_count[self._camera_name]}"
    self._framerate = displayer_framerate
    self._displayer_backend = displayer_backend

  def __del__(self) -> None:
    """Safety method called when deleting the Block and ensuring that all the
    instantiated :class:`~crappy.blocks.camera_processes.CameraProcess` as well 
    as the :obj:`~multiprocessing.Manager` are stopped before exiting.
    
    If they did not stop in time, just terminates them.
    """

    if self.process_proc is not None and self.process_proc.is_alive():
      self.process_proc.terminate()

    if self._save_proc is not None and self._save_proc.is_alive():
      self._save_proc.terminate()

    if self._display_proc is not None and self._display_proc.is_alive():
      self._display_proc.terminate()

    if self._manager is not None:
      self._manager.shutdown()

  def prepare(self) -> None:
    """Preparing the save folder, opening the camera and displaying the
    configuration GUI.
    
    This method calls the :meth:`crappy.camera.Camera.open` method of the
    :class:`~crappy.camera.Camera` object.
    """

    # Instantiating the synchronization objects
    self.log(logging.DEBUG, "Instantiating the multiprocessing "
                            "synchronization objects")
    self._manager = Manager()
    self._metadata = self._manager.dict()
    self._stop_event_cam = Event()
    self._overlay_conn_in, self._overlay_conn_out = Pipe()
    self._save_lock = RLock()
    self._disp_lock = RLock()
    self._proc_lock = RLock()

    # instantiating the ImageSaver CameraProcess
    if self._save_images:
      self.log(logging.INFO, "Instantiating the saver process")
      self._save_proc = ImageSaver(img_extension=self._img_extension,
                                   save_folder=self._save_folder,
                                   save_period=self._save_period,
                                   save_backend=self._save_backend)

    # instantiating the Displayer CameraProcess
    if self._display_images:
      self.log(logging.INFO, "Instantiating the displayer process")
      self._display_proc = Displayer(title=self._title,
                                     framerate=self._framerate,
                                     backend=self._displayer_backend)

    # Creating the Barrier for the synchronization of the CameraProcesses
    n_proc = sum(int(proc is not None) for proc in (self.process_proc,
                                                    self._save_proc,
                                                    self._display_proc))
    if not n_proc:
      self.log(logging.WARNING, "The Block acquires images but does not save "
                                "them, nor display them, nor process them !")

    self._cam_barrier = Barrier(n_proc + 1)

    # Case when the images are artificially generated and not acquired
    if self._image_generator is not None:
      self.log(logging.INFO, "Setting the image generator camera")
      self._camera = BaseCam()
      self._camera.add_scale_setting('Exx', -100., 100., None, None, 0.)
      self._camera.add_scale_setting('Eyy', -100., 100., None, None, 0.)
      img = self._image_generator(0, 0)
      self._camera.add_software_roi(img.shape[1], img.shape[0])
      self._camera.set_all()

      def get_image(self_) -> (float, np.ndarray):
        """Method generating the frames using the ``image_generator`` argument 
        if one was provided."""
        
        return time(), self_.apply_soft_roi(self._image_generator(self_.Exx,
                                                                  self_.Eyy))

      self._camera.get_image = MethodType(get_image, self._camera)

    # Instantiating the Camera object for acquiring the images
    else:
      self._camera = camera_dict[self._camera_name]()
      self.log(logging.INFO, f"Opening the {self._camera_name} Camera")
      self._camera.open(**self._camera_kwargs)
      self.log(logging.INFO, f"Opened the {self._camera_name} Camera")

    # Displaying the configuration window if required
    if self._config_cam:
      self.log(logging.INFO, "Displaying the configuration window")
      self._configure()
      self.log(logging.INFO, "Camera configuration done")

    # Setting the camera to 'Hardware' trig if it's in 'Hdw after config' mode
    if self._camera.trigger_name in self._camera.settings and \
        getattr(self._camera,
                self._camera.trigger_name) == 'Hdw after config':
      self.log(logging.INFO, "Setting the trigger mode to Hardware")
      setattr(self._camera, self._camera.trigger_name, 'Hardware')

    # Ensuring a dtype and a shape were given for the image
    if self._img_dtype is None or self._img_shape is None:
      raise ValueError(f"Cannot launch the Camera processes for camera "
                       f"{self._camera_name} as the image shape and/or dtype "
                       f"wasn't specified.\n Please specify it in the args, or"
                       f" enable the configuration window.")

    # Instantiating the Array for sharing the frames with the CameraProcesses
    self.log(logging.DEBUG, "Instantiating the shared objects")
    self._img_array = Array(np.ctypeslib.as_ctypes_type(self._img_dtype),
                            int(np.prod(self._img_shape)))
    self._img = np.frombuffer(self._img_array.get_obj(),
                              dtype=self._img_dtype).reshape(self._img_shape)

    # Starting the CameraProcess for image processing if it was instantiated
    if self.process_proc is not None:
      self.log(logging.DEBUG, "Sharing the synchronization objects with the "
                              "image processing process")
      overlay_conn = (self._overlay_conn_in if self._display_proc is not None
                      else None)
      labels = self.labels if self.labels is not None else None
      self.process_proc.set_shared(array=self._img_array,
                                   data_dict=self._metadata,
                                   lock=self._proc_lock,
                                   barrier=self._cam_barrier,
                                   event=self._stop_event_cam,
                                   shape=self._img_shape,
                                   dtype=self._img_dtype,
                                   to_draw_conn=overlay_conn,
                                   outputs=self.outputs,
                                   labels=labels,
                                   log_queue=self._log_queue,
                                   log_level=self._log_level,
                                   display_freq=self.display_freq)
      self.log(logging.INFO, "Starting the image processing process")
      self.process_proc.start()

    # Starting the ImageSaver CameraProcess if it was instantiated
    if self._save_proc is not None:
      self.log(logging.DEBUG, "Sharing the synchronization objects with the "
                              "image saver process")
      self._save_proc.set_shared(array=self._img_array,
                                 data_dict=self._metadata,
                                 lock=self._save_lock,
                                 barrier=self._cam_barrier,
                                 event=self._stop_event_cam,
                                 shape=self._img_shape,
                                 dtype=self._img_dtype,
                                 to_draw_conn=None,
                                 outputs=list(),
                                 labels=list(),
                                 log_queue=self._log_queue,
                                 log_level=self._log_level,
                                 display_freq=self.display_freq)
      self.log(logging.INFO, "Starting the image saver process")
      self._save_proc.start()

    # Starting the Displayer CameraProcess if it was instantiated
    if self._display_proc is not None:
      self.log(logging.DEBUG, "Sharing the synchronization objects with the "
                              "image displayer process")
      self._display_proc.set_shared(array=self._img_array,
                                    data_dict=self._metadata,
                                    lock=self._disp_lock,
                                    barrier=self._cam_barrier,
                                    event=self._stop_event_cam,
                                    shape=self._img_shape,
                                    dtype=self._img_dtype,
                                    to_draw_conn=self._overlay_conn_out,
                                    outputs=list(),
                                    labels=list(),
                                    log_queue=self._log_queue,
                                    log_level=self._log_level,
                                    display_freq=self.display_freq)
      self.log(logging.INFO, "Starting the image displayer process")
      self._display_proc.start()

  def begin(self) -> None:
    """This method waits for all the 
    :class:`~crappy.blocks.camera_processes.CameraProcess` to be ready, then
    releases them all at once to make sure they're synchronized.
    
    
    A :obj:`~multiprocessing.Barrier` is used for forcing the CameraProcesses
    to wait for each other.
    """

    try:
      self.log(logging.INFO, "Waiting for all Camera processes to be ready")
      self._cam_barrier.wait()
      self.log(logging.INFO, "All Camera processes ready now")
    except BrokenBarrierError:
      raise CameraPrepareError

    self._last_cam_fps = time()

  def loop(self) -> None:
    """This method receives data from upstream Blocks, acquires a frame from 
    the :class:`~crappy.camera.Camera` object, and transmits it to all the 
    :class:`~crappy.blocks.camera_processes.CameraProcess`.

    The image is acquired by calling the 
    :meth:`~crappy.camera.Camera.get_image` method of the Camera object. If
    only a timestamp is returned by this method, and not a complete :obj:`dict`
    of metadata, some basic metadata is generated here and transmitted to the
    CameraProcesses.
    
    This method also manages the software trigger if this option was set, 
    applies the image transformation function if one was given, and displays
    the FPS of the acquisition if required.
    """

    # Signaling all the Blocks to stop if a CameraProcess crashed
    if self._stop_event_cam.is_set():
      raise CameraRuntimeError

    # Receiving the data from upstream Blocks
    data = self.recv_last_data(fill_missing=False)

    # Waiting for the trig label if one was given
    if self._trig_label is not None and self._trig_label not in data:
      return
    elif self._trig_label is not None and self._trig_label in data:
      self.log(logging.DEBUG, "Software trigger signal received")

    # Updating the image generator if one was provided
    if self._image_generator is not None:
      if 'Exx(%)' in data:
        self.log(logging.DEBUG, f"Setting Exx to {data['Exx(%)']}")
        self._camera.Exx = data['Exx(%)']
      if 'Eyy(%)' in data:
        self.log(logging.DEBUG, f"Setting Eyy to {data['Eyy(%)']}")
        self._camera.Eyy = data['Eyy(%)']

    # Grabbing the frame from the Camera object
    ret = self._camera.get_image()
    if ret is None:
      return
    metadata, img = ret
 
    # Building the metadata dict if it was not provided
    if isinstance(metadata, float):
      metadata = {'t(s)': metadata,
                  'DateTimeOriginal': strftime("%Y:%m:%d %H:%M:%S",
                                               gmtime(metadata)),
                  'SubsecTimeOriginal': f'{metadata % 1:.6f}',
                  'ImageUniqueID': self._loop_count}

    # Making the timestamp relative to the beginning of the test
    metadata['t(s)'] -= self.t0

    # Applying the transform function if one as provided
    if self._transform is not None:
      img = self._transform(img)

    # Copying the metadata and the acquired frame into the shared objects for 
    # transfer to the CameraProcesses
    # This is done with all the Locks acquired to avoid any conflict
    with self._save_lock, self._disp_lock, self._proc_lock:
      self.log(logging.DEBUG, f"Writing metadata to shared dict: {metadata}")
      self._metadata.update(metadata)
      self.log(logging.DEBUG, "Writing image to shared array")
      np.copyto(self._img, img)

    self._loop_count += 1

    # If requested, displays the FPS of the image acquisition
    if self.display_freq:
      self._fps_count += 1
      t = time()
      if t - self._last_cam_fps > 2:
        self.log(logging.INFO, f"Acquisition FPS: "
                               f"{self._fps_count / (t - self._last_cam_fps)}")
        self._last_cam_fps = t
        self._fps_count = 0

  def finish(self) -> None:
    """This method stops the image acquisition on the 
    :class:`~crappy.camera.Camera`, as well as all the 
    :class:`~crappy.blocks.camera_processes.CameraProcess` that were started.
    
    If the CameraProcesses do not gently stop, they are terminated. Also stops
    the :obj:`~multiprocessing.Manager` in charge of handling the metadata.
    
    For stopping the image acquisition, the :meth:`~crappy.camera.Camera.close`
    method is called.
    """

    # Closing the Camera object
    if self._image_generator is None and self._camera is not None:
      self.log(logging.INFO, f"Closing the {self._camera_name} Camera")
      self._camera.close()
      self.log(logging.INFO, f"Closed the {self._camera_name} Camera")

    # Setting the stop event to signal all CameraProcesses to stop
    if self._stop_event_cam is not None:
      self.log(logging.DEBUG, "Asking all the children processes to stop")
      self._stop_event_cam.set()
      sleep(0.2)

    # If the processing CameraProcess is not done, terminating it
    if self.process_proc is not None and self.process_proc.is_alive():
      self.log(logging.WARNING, "Image processing process not stopped, "
                                "killing it !")
      self.process_proc.terminate()
    # If the ImageSaver CameraProcess is not done, terminating it
    if self._save_proc is not None and self._save_proc.is_alive():
      self.log(logging.WARNING, "Image saver process not stopped, "
                                "killing it !")
      self._save_proc.terminate()
    # If the Displayer CameraProcess is not done, terminating it
    if self._display_proc is not None and self._display_proc.is_alive():
      self.log(logging.WARNING, "Image displayer process not stopped, "
                                "killing it !")
      self._display_proc.terminate()

    # Closing the Manager handling the metadata
    if self._manager is not None:
      self._manager.shutdown()

  def _configure(self) -> None:
    """This method should instantiate and start the 
    :class:`~crappy.tool.camera_config.CameraConfig` window for configuring the
    :class:`~crappy.camera.Camera` object.
    
    It should also handle the case when an exception is raised in the 
    configuration window.
    
    This method is meant to be overriden by children of the Camera Block, as 
    other image processing Blocks rely on subclasses of 
    :class:`~crappy.tool.camera_config.CameraConfig`.
    """

    config = None
    
    # Instantiating and starting the configuration window
    try:
      config = CameraConfig(self._camera, self._log_queue,
                            self._log_level, self.freq)
      config.main()
    
    # If an exception is raised in the config window, closing it before raising
    except (Exception,) as exc:
      self._logger.exception("Caught exception in the configuration window !",
                             exc_info=exc)
      if config is not None:
        config.stop()
      raise CameraConfigError

    # Getting the image dtype and shape for setting the shared Array
    if config.shape is not None:
      self._img_shape = config.shape
    if config.dtype is not None:
      self._img_dtype = config.dtype



# coding: utf-8

from __future__ import annotations
from datetime import timedelta
from time import time
from typing import Tuple, Dict, Any, Optional, Iterable
import logging

from .meta_block import Block
from .._global import OptionalModule

plt = OptionalModule('matplotlib.pyplot', lazy_import=True)
mpl = OptionalModule('matplotlib', lazy_import=True)


class Text:
  """Displays a simple text line on the drawing.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               _: Canvas,
               coord: Tuple[int, int],
               text: str,
               label: str,
               **__: str) -> None:
    """Sets the arguments.

    Args:
      _: The parent drawing Block.
      coord: The coordinates of the text on the drawing.
      text: The text to display.
      label: The label carrying the information for updating the text.
      **__: Other unused arguments.

    .. versionchanged:: 1.5.10
       now explicitly listing the *_*, *coord*, *text* and *label* arguments
    """

    x, y = coord
    self._text = text
    self._label = label

    self._txt = plt.text(x, y, text)

  def update(self, data: Dict[str, float]) -> None:
    """Updates the text according to the received values."""

    if self._label in data:
      self._txt.set_text(self._text % data[self._label])


class DotText:
  """Like :class:`Text`, but with a colored dot to visualize a numerical value.

  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Dot_text to DotText
  """

  def __init__(self,
               drawing: Canvas,
               coord: Tuple[int, int],
               text: str,
               label: str,
               **__: str) -> None:
    """Sets the arguments.

    Args:
      drawing: The parent drawing Block.
      coord: The coordinates of the text and the color dot on the drawing.
      text: The text to display.
      label: The label carrying the information for updating the text and the
        color of the dot.
      **__: Other unused arguments.

    Important:
      The value received in label must be a numeric value. It will be
      normalized on the ``crange`` of the Block and the dot will change
      color from blue to red depending on this value.
      
    .. versionchanged:: 1.5.10
       now explicitly listing the *drawing*, *coord*, *text* and *label*
       arguments
    """

    x, y = coord
    self._text = text
    self._label = label

    self._txt = plt.text(x + 40, y + 20, text, size=16)
    self._dot = plt.Circle(coord, 20)

    drawing.ax.add_artist(self._dot)
    low, high = drawing.color_range

    self._amp = high - low
    self._low = low

  def update(self, data: Dict[str, float]) -> None:
    """Updates the text and the color dot according to the received values."""

    if self._label in data:
      self._txt.set_text(self._text % data[self._label])
      self._dot.set_color(mpl.cm.coolwarm((data[self._label] -
                                           self._low) / self._amp))


class Time:
  """Displays a time counter on the drawing, starting at the beginning of the
  test.

  .. versionadded:: 1.4.0
  """

  def __init__(self, drawing: Canvas, coord: Tuple[int, int], **__) -> None:
    """Sets the arguments.

    Args:
      drawing: The parent drawing Block.
      coord: The coordinates of the time counter on the drawing.
      **__: Other unused arguments.

    .. versionchanged:: 1.5.10
       now explicitly listing the *drawing* and *coord* arguments
    """

    self._block = drawing
    x, y = coord

    self._txt = plt.text(x, y, "00:00", size=38)

  def update(self, _: Dict[str, float]) -> None:
    """Updates the time counter, independently of the received values."""

    self._txt.set_text(str(timedelta(seconds=int(time() - self._block.t0))))


class Canvas(Block):
  """This Block allows displaying a real-time visual representation of data.

  It displays the data on top of a background image and updates it according to
  the values received through the incoming :class:`~crappy.links.Link`. The
  background image and the data overlay are displayed in a new window.

  It is possible to display a simple text, a time counter, or text associated
  with a color dot evolving depending on a predefined color bar and the
  received values.

  This Block is mostly useful for displaying a user-friendly and fine-tuned
  representation of data. For simpler displays, the
  :class:`~crappy.blocks.Dashboard`, :class:`~crappy.blocks.Grapher` and
  :class:`~crappy.blocks.LinkReader` Blocks should be preferred.

  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Drawing to Canvas
  """

  def __init__(self,
               image_path: str,
               draw: Optional[Iterable[Dict[str, Any]]] = None,
               color_range: Tuple[float, float] = (20, 300),
               title: str = "Canvas",
               window_size: Tuple[int, int] = (7, 5),
               backend: str = "TkAgg",
               freq: Optional[float] = 2,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      image_path: Path to the image that will be the background of the canvas,
        as a :obj:`str`.
      draw: An iterable (like a :obj:`list` or a :obj:`tuple`) of :obj:`dict`
        defining what to draw. See below for more details.
      color_range: A :obj:`tuple` containing the lowest and highest values for
        the color bar.
      title: The title of the window containing the drawing.
      window_size: The `x` and `y` dimension of the window, following
        :mod:`matplotlib` nomenclature.
      backend: The :mod:`matplotlib` backend to use.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

    Note:
      - Information about the ``draw`` keys:

        - ``type``: Mandatory, the type of drawing to display. It can be either
          `'text'`, `'dot_text'` or `'time'`.

        - ``coord``: Mandatory, a :obj:`tuple` containing the `x` and `y`
          coordinates where the element should be displayed on the drawing.

        - ``text``: Mandatory for `'text'` and `'dot_text'` only, the text to
          display on the drawing. It must follow the %-formatting, and contain
          exactly one %-field. Ex: `'T0 = %f'`. This field will be updated
          using the value carried by ``label``.

        - ``label``: Mandatory for `'text'` and `'dot_text'` only, the label of
          the data to display. It will try to retrieve this data in the
          incoming Links. The ``text`` will then be updated with this data.
    
    .. versionchanged:: 1.5.10 renamed *crange* argument to *color_range*
    .. versionadded:: 1.5.10 *verbose* argument
    .. versionadded:: 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._image = image_path
    self._draw = [] if draw is None else list(draw)
    self.color_range = color_range
    self._title = title
    self._window_size = window_size
    self._backend = backend

    self._fig = None
    self.ax = None
    self._drawing_elements = None

  def prepare(self) -> None:
    """Initializes the different elements of the drawing."""

    self.log(logging.INFO, "Opening the drawing windows")

    # Initializing the window and the background image
    plt.switch_backend(self._backend)
    self._fig, self.ax = plt.subplots(figsize=self._window_size)
    image = self.ax.imshow(plt.imread(self._image), cmap=mpl.cm.coolwarm)
    image.set_clim(-0.5, 1)

    # Initializing the color bar
    cbar = self._fig.colorbar(image, ticks=[-0.5, 1], fraction=0.061,
                              orientation='horizontal', pad=0.04)
    cbar.set_label('Dot text values')
    cbar.ax.set_xticklabels(self.color_range)

    # Setting the title and the axes
    self.ax.set_title(self._title)
    self.ax.set_axis_off()

    # Adding the elements to the drawing
    self._drawing_elements = []
    for dic in self._draw:
      if dic['type'] == 'text':
        self._drawing_elements.append(Text(self, **dic))
      elif dic['type'] == 'dot_text':
        self._drawing_elements.append(DotText(self, **dic))
      elif dic['type'] == 'time':
        self._drawing_elements.append(Time(self, **dic))

  def loop(self) -> None:
    """Receives the latest data from upstream Blocks and updates the drawing
    accordingly."""

    data = self.recv_last_data(fill_missing=False)
    if not data:
      return

    for elt in self._drawing_elements:
      elt.update(data)
    self.log(logging.DEBUG, "Updating the drawing window")
    self._fig.canvas.draw()
    plt.pause(0.001)

  def finish(self) -> None:
    """Closes the window containing the drawing."""

    self.log(logging.INFO, "Closing the drawing windows")
    plt.close()



# coding: utf-8

from time import time, sleep
from re import fullmatch, findall
from typing import List, Union, Optional, Iterable
import logging
from  warnings import warn

from ..meta_inout import InOut
from ...tool.ft232h import FT232HServer as FT232H, USBArgsType

# ADS1256 gain channel
Ads1256_gain = {1: 0b000,
                2: 0b001,
                4: 0b010,
                8: 0b011,
                16: 0b100,
                32: 0b101,
                64: 0b110}

# ADS1256 data rate
Ads1256_drate = {2.5: 0x03,
                 5: 0x13,
                 10: 0x20,
                 15: 0x33,
                 25: 0x43,
                 30: 0x53,
                 50: 0x63,
                 60: 0x72,
                 100: 0x82,
                 500: 0x92,
                 1000: 0xA1,
                 2000: 0xB0,
                 3750: 0xC0,
                 7500: 0xD0,
                 15000: 0xE0,
                 30000: 0xF0}

# Ads1256 registration definition
Ads1256_reg = {'REG_STATUS': 0,
               'REG_MUX': 1,
               'REG_ADCON': 2,
               'REG_DRATE': 3,
               'REG_IO': 4,
               'REG_OFC0': 5,
               'REG_OFC1': 6,
               'REG_OFC2': 7,
               'REG_FSC0': 8,
               'REG_FSC1': 9,
               'REG_FSC2': 10}

# Ads1256 command definition
Ads1256_cmd = {'CMD_WAKEUP': 0x00,
               'CMD_RDATA': 0x01,
               'CMD_RDATAC': 0x03,
               'CMD_SDATAC': 0x0F,
               'CMD_RREG': 0x10,
               'CMD_WREG': 0x50,
               'CMD_SELFCAL': 0xF0,
               'CMD_SELFOCAL': 0xF1,
               'CMD_SELFGCAL': 0xF2,
               'CMD_SYSOCAL': 0xF3,
               'CMD_SYSGCAL': 0xF4,
               'CMD_SYNC': 0xFC,
               'CMD_STANDBY': 0xFD,
               'CMD_RESET': 0xFE}

# Dac8532 channels definition
Dac8532_chan = {0: 0x10,
                1: 0x24}


class WaveshareADDAFT232H(InOut):
  """Class for controlling Waveshare's AD/DA Raspberry Pi hat through an
  FT232H.

  It is similar to the :class:`~crappy.inout.WaveshareADDA` class, except this
  class is specific for use with an :class:`~crappy.tool.ft232h.FT232H` USB to
  I2C converter.

  It communicates over the SPI protocol and the GPIOs. It allows to
  read values from the 8-channels ADC and/or to set the 2-channels DAC. The hat
  can acquire up to 30000 samples per second, although this data rate is
  impossible to achieve using Crappy.

  .. versionadded:: 2.0.0
  """

  ft232h = True

  def __init__(self,
               dac_channels: Optional[Iterable[str]] = None,
               adc_channels: Optional[Iterable[str]] = None,
               gain_hardware: int = 1,
               v_ref: float = 3.3,
               gain: float = 1,
               offset: float = 0,
               sample_rate: Union[int, float] = 100,
               _ft232h_args: USBArgsType = tuple(),
               rst_pin_ads: str = 'D7',
               cs_pin_ads: str = 'D4',
               drdy_pin_ads: str = 'D6',
               cs_pin_dac: str = 'D5') -> None:
    """Checks the validity of the arguments.

    Args:
      dac_channels: An iterable (like a :obj:`list` or a :obj:`tuple`) of
        :obj:`str` representing the channels to be set. The syntax for each
        string is 'DACi' with i being either `0` or `1`.
      adc_channels: An iterable (like a :obj:`list` or a :obj:`tuple`) of
        :obj:`str` representing the channels to read. The syntax for all the
        strings is either:
        ::

          'ADi' (i in range(8))

        or else:
        ::

          'ADi - ADj' (i, j in range(8))

      gain_hardware: The gain to be used by the programmable gain amplifier.
        Setting a high gain allows to read small voltages with a better
        precision, but it might saturate the sensor for higher voltages. The
        available gain values are:
        ::

          1, 2, 4, 8, 16, 32, 64

      v_ref: The voltage reference set by the `VREF` jumper. When reading
        single inputs, ``v_ref`` is the value the ADC compares the signals
        with. In a similar way, the maximum output voltage of the DAC is
        ``v_ref``. `3.3` and `5` are the only possible values for this setting,
        as the FT232H can only provide `3.3V` and `5V`.
      gain: Allows to tune the output values of the ADC according to the
        formula : :math:`output = gain * tension + offset`. The same gain
        applies to all the outputs.
      offset: Allows to tune the output values of the ADC according to the
        formula : :math:`output = gain * tension + offset`. The same offset
        applies to all the outputs.
      sample_rate: The ADC data output rate in SPS. The available values are:
        ::

          2.5, 5, 10, 15, 25, 30, 50, 60, 100, 500,
          1000, 2000, 3750, 7500, 15000, 30000

      _ft232h_args: This argument is meant for internal use only and should not
        be provided by the user. It contains the information necessary for
        setting up the FT232H.
      rst_pin_ads: The pin for resetting the ADS1256.
      cs_pin_ads: The chip select pin for the ADS1256.
      drdy_pin_ads: The pin for knowing when a conversion result in ready.
      cs_pin_dac: The chip select pin for the DAC8552.

    Warning:
      - ``adc_channels``:
        For reading single inputs the `JMP_AGND` jumper should normally be
        connected, whereas it should be disconnected for reading differential
        inputs. It is however possible to set a different reference than `AGND`
        for single input measurements, in which case the `JMP_AGND` jumper
        should not be connected and the voltage reference should be plugged
        in the `AINCOM` pin.

        The AD/DA offers the possibility to read single inputs or
        differential inputs, but not both at the same time ! This is due to
        the `JMP_AGND` jumper.
        For measuring both input types simultaneously, is it necessary to
        connect `AGND` to one of the channels (for example `AD0`). Then all
        single inputs `'ADi'` should be replaced by `'ADi - AD0'`. They are
        then considered as differential inputs.

        The ADC channels voltages should not be lower than `AGND-0.1V`, and not
        be greater than `AGND+5.1V`. This is independent of `VREF` value.

    Note:
      - ``adc_channels``:
        If multiple channels to read are given, they are read in a sequential
        way. This means that there's a small delay between each acquisition,
        and the timeframe is thus less accurate for the last channels than
        for the first ones. If time precision matters it is preferable to
        read as few channels as possible !

      - ``vref``:
        `VREF` can be set independently of the chosen `VCC` value. The `VCC`
        value has no influence on the ADC behaviour as it is always powered
        up with `5V`. Same goes for the DAC.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._bus = None

    super().__init__()

    (block_index, block_lock, command_file, answer_file, shared_lock,
     current_block) = _ft232h_args

    if gain_hardware not in Ads1256_gain:
      raise ValueError("gain_hardware should be in {}".format(list(
        Ads1256_gain.keys())))
    else:
      self._gain_hardware = gain_hardware

    if sample_rate not in Ads1256_drate:
      raise ValueError("sample_rate should be in {}".format(list(
        Ads1256_drate.keys())))
    else:
      self._sample_rate = sample_rate

    if v_ref not in [3.3, 5]:
      raise ValueError("v_ref should be either 3.3 or 5")
    else:
      self._v_ref = v_ref

    self._channel_set = False

    if dac_channels is None and adc_channels is None:
      self.log(logging.WARNING, "The AD/DA doesn't read nor write anything")

    self._channels_write = []
    if dac_channels is not None:
      for chan in dac_channels:
        # Checking if the format matches
        if fullmatch(r'DAC[0-1]', chan) is None:
          raise ValueError("Valid format for dac_channels values is 'DACi' "
                           "with i either 0 or 1")
        else:
          self._channels_write.append(int(findall(r'\d', chan)[0]))

      self.log(logging.DEBUG, f"Channels to write to: {self._channels_write}")

    self._channels_read = []
    if adc_channels is not None:
      for chan in adc_channels:
        # Checking if the format matches
        if fullmatch(r'AD[0-7]', chan) is None and \
              fullmatch(r'AD[0-7]\s?-\s?AD[0-7]', chan) is None:
          raise ValueError("Valid formats for adc_channels values are "
                           "either 'ADi' (i in range(8)) or 'ADi - ADj'")
        else:
          # Extracting the channel numbers
          chan = [int(i) for i in findall(r'\d', chan)]
          # If only one channel provided, it has to be compared to GND
          if len(chan) == 1:
            chan.append(8)
          # Making sure the two channels are not identical
          if chan[0] == chan[1]:
            raise ValueError("The two channels are the same !")
          self._channels_read.append(chan)

      self.log(logging.DEBUG, f"Channels to read from: {self._channels_read}")

    self._gain = gain
    self._offset = offset

    self.log(logging.INFO, "Opening the SPI communication with the AD/DA")
    self._bus = FT232H(mode='SPI',
                       block_index=block_index,
                       current_block=current_block,
                       command_file=command_file,
                       answer_file=answer_file,
                       block_lock=block_lock,
                       shared_lock=shared_lock)
    self._rst_pin_ads = rst_pin_ads
    self._cs_pin_ads = cs_pin_ads
    self._drdy_pin_ads = drdy_pin_ads
    self._cs_pin_dac = cs_pin_dac

  def open(self) -> None:
    """Sets the SPI communication, the GPIOs and the device."""

    # Setting the SPI
    self.log(logging.INFO, "Setting up the SPI connection")
    self._bus.max_speed_hz = 40000
    self._bus.mode = 1
    self._bus.no_cs = True
    self._reset()

    # Setting the ADS according to the user parameters
    buf = [Ads1256_gain[self._gain_hardware], Ads1256_drate[self._sample_rate]]
    self._bus.set_gpio(self._cs_pin_ads, False)
    cmd = [Ads1256_cmd['CMD_WREG'] | Ads1256_reg['REG_ADCON'], 0x01] + buf
    self.log(logging.DEBUG, f"Writing {cmd} to the SPI bus")
    self._bus.writebytes(cmd)
    self._bus.set_gpio(self._cs_pin_ads, True)
    sleep(0.001)

  def get_data(self) -> List[float]:
    """Reads data from all the user-specified ADC channels, in a sequential
    way.

    Data is returned in Volts, but this can be tuned using gain and offset.

    Returns:
      A :obj:`list` containing the timestamp, and then the values for each
      channel to read.
    """

    out = [time()]

    # The values are read one channel after the other, not simultaneously
    for chan in self._channels_read:
      self._bus.set_gpio(self._cs_pin_ads, False)

      # Switching channel only if necessary, except for the first loop
      if len(self._channels_read) > 1 or not self._channel_set:
        cmd = [Ads1256_cmd['CMD_WREG'] | Ads1256_reg['REG_MUX'], 0x00,
               (chan[0] << 4) | chan[1]]
        self.log(logging.DEBUG, f"Writing {cmd} to the SPI bus")
        self._bus.writebytes(cmd, stop=False)

        # The ADS has to be synchronized again when switching channel
        self.log(logging.DEBUG, f"Writing {[Ads1256_cmd['CMD_SYNC']]} to the "
                                f"SPI bus")
        self._bus.writebytes([Ads1256_cmd['CMD_SYNC']],
                             start=False,
                             stop=False)
        self.log(logging.DEBUG, f"Writing {[Ads1256_cmd['CMD_WAKEUP']]} to "
                                f"the SPI bus")
        self._bus.writebytes([Ads1256_cmd['CMD_WAKEUP']], start=False)

        self._channel_set = True

      # Reading the output value
      self._wait_drdy()
      self.log(logging.DEBUG, f"Writing {[Ads1256_cmd['CMD_RDATA']]} to the "
                              f"SPI bus")
      self._bus.writebytes([Ads1256_cmd['CMD_RDATA']], stop=False)
      buf = self._bus.readbytes(3, start=False)
      self.log(logging.DEBUG, f"Read {buf} from the SPI bus")
      self._bus.set_gpio(self._cs_pin_ads, True)

      # Converting the raw output into Volts
      out_raw = (buf[0] << 16) | (buf[1] << 8) | buf[2]
      if out_raw & 0x800000:
        out_raw -= 2 ** 24
      value = self._v_ref / self._gain_hardware * out_raw / 2 ** 23
      out.append(self._offset + self._gain * value)

    return out

  def set_cmd(self, *cmd: float) -> None:
    """Sets the user-specified DAC channels according to the input values.

    Args:
      cmd: The input values as :obj:`float`, in Volts.
    """

    # The values are set one channel after the other, not simultaneously
    for val, channel in zip(cmd, self._channels_write):
      if not 0 <= val <= self._v_ref:
        raise ValueError("Desired output voltage should be between 0 and "
                         "v_ref")
      digit = int((2 ** 16 - 1) * val / self._v_ref)
      self._bus.set_gpio(self._cs_pin_dac, False)
      cmd = [Dac8532_chan[channel], digit >> 8, digit & 0xFF]
      self.log(logging.DEBUG, f"Writing {cmd} to the SPI bus")
      self._bus.writebytes(cmd)
      self._bus.set_gpio(self._cs_pin_dac, True)

  def close(self) -> None:
    """Releases the GPIOs."""

    if self._bus is not None:
      self.log(logging.INFO, "Closing the SPI communication with the AD/DA")
      self._bus.close()

  def _reset(self) -> None:
    """Resets the ADC."""

    self._bus.set_gpio(self._cs_pin_ads, True)
    sleep(0.2)
    self._bus.set_gpio(self._cs_pin_ads, False)
    sleep(0.2)
    self._bus.set_gpio(self._cs_pin_ads, True)

  def _wait_drdy(self) -> None:
    """Waits until the `DRDY` pin is driven low, meaning that an ADC conversion
    is ready."""

    t0 = time()
    while self._bus.get_gpio(self._drdy_pin_ads):
      if time() - t0 > 1:
        raise TimeoutError("Couldn't get conversion result from the ADC")



# coding: utf-8

from struct import pack_into
from time import sleep
from typing import Union, Tuple
import logging
from  warnings import warn

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  from adafruit_motorkit import MotorKit
except (ImportError, ModuleNotFoundError):
  MotorKit = OptionalModule('adafruit_motorkit',
                            'Adafrfuit Motorkit module (adafruit_motorkit) is '
                            'required to use this actuator')

try:
  import board
except (ImportError, ModuleNotFoundError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  from smbus2 import SMBus
except (ImportError, ModuleNotFoundError):
  SMBus = OptionalModule('smbus2')

motor_hat_ctrl = {1: 0x26,
                  2: 0x3A,
                  3: 0x0E,
                  4: 0x22}

motor_hat_pos = {1: 0x2A,
                 2: 0x32,
                 3: 0x12,
                 4: 0x1A}

motor_hat_neg = {1: 0x2E,
                 2: 0x36,
                 3: 0x16,
                 4: 0x1E}

motor_hat_0xFF = [0x00, 0x10, 0x00, 0x00]
motor_hat_backends = ['Pi4', 'blinka']
motor_hat_max_volt = 12


class DCMotorHat(Actuator):
  """Class for driving Adafruit's DC motor HAT.

  It can drive up to four DC motors in speed only. The acquisition of the speed
  has not been implemented so far. It can either rely on Adafruit's Blinka
  library, or on :mod:`smbus2` if used from a Raspberry Pi.

  Important:
    As this Actuator can drive up to 4 motors simultaneously, it takes a
    :obj:`tuple` as a command, see :meth:`set_speed`. Regular Actuators receive
    their commands as :obj:`float`. A :class:`~crappy.modifier.Modifier` can be
    used for converting a :obj:`float` command from a
    :class:`~crappy.blocks.Generator` to a :obj:`tuple`.

  Note:
    The DC Motor Hat can also drive stepper motors, but this feature isn't
    included here.
  
  .. versionadded:: 2.0.0
  """

  def __init__(self,
               backend: str,
               device_address: int = 0x60,
               i2c_port: int = 1) -> None:
    """Checks the validity of the arguments.

    Args:
      backend: Should be one of :
        ::

          'Pi4', 'blinka'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing
        :mod:`adafruit-circuitpython-motorkit` and :mod:`Adafruit-Blinka`, but
        these modules are compatible with and maintained on a wide variety of
        boards.
      device_address: The I2C address of the HAT. The default address is
        `0x60`, but it is possible to change this setting by cutting traces on
        the board.
      i2c_port: The I2C port over which the HAT should communicate. On most
        Raspberry Pi models the default I2C port is `1`.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._bus = None
    self._buf = bytearray(4)

    if not isinstance(backend, str) or backend not in motor_hat_backends:
      raise ValueError("backend should be in {}".format(motor_hat_backends))
    self._backend = backend

    super().__init__()

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer between 0 and 127.")
    self._address = device_address

    if not isinstance(i2c_port, int):
      raise TypeError("i2c_port should be an integer !")
    self._port = i2c_port

  def open(self) -> None:
    """Opens the connection to the motor hat and initializes it."""

    if self._backend == 'blinka':
      self.log(logging.INFO, "Opening with the Adafruit library")
      self._bus = MotorKit(i2c=board.I2C())

    elif self._backend == 'Pi4':
      self.log(logging.INFO, "Opening with the SMBus library")

      self._bus = SMBus(self._port)

      # Reset
      self._write_i2c(0x00, [0x00])
      sleep(0.01)

      # Sleep & restart, also setting the frequency to 1526Hz
      self._write_i2c(0x00, [0x10])
      self._write_i2c(0xFE, [int(25E6 / 4096.0 / 1600.0 + 0.5)])
      sleep(0.01)
      self._write_i2c(0x00, [0x00])
      sleep(0.01)
      self._write_i2c(0x00, [0xA0])

      # Initializing the motors
      for i in range(1, 5):
        self._write_i2c(motor_hat_ctrl[i], motor_hat_0xFF)

  def set_speed(self, cmd: Tuple[int, float]) -> None:
    """Sets the desired voltage on the selected motor.

    The provided voltage should be between `-12` and `12V` which are the limits
    of the motor hat. If not, it will be silently clamped in this interval.

    Warning:
      Unlike most Actuators, this one takes :obj:`tuple` as commands instead
      of :obj:`float`.

    Args:
      cmd: A :obj:`tuple` containing first the index of the motor to drive as
        an :obj:`int` between 1 and 5, and second the voltage to apply to the
        selected motor as a :obj:`float`.
    """

    try:
      motor_nr, volts = cmd
    except (TypeError, ValueError):
      raise ValueError("The DCMotorHat receives commands as tuples, with first"
                       " the index of the motor and second the voltage "
                       "command")

    if motor_nr not in range(1, 5):
      raise ValueError("The index of the motor to drive should be an integer "
                       "between 1 and 5")

    volt_clamped = min(abs(volts) / motor_hat_max_volt, 1.0)
    self.log(logging.DEBUG, f"Setting motor {motor_nr} to {volt_clamped}V")
    self._set_motor(motor_nr, volt_clamped)

  def stop(self) -> None:
    """Simply sets the command to `0` to stop the motors."""

    if self._bus is not None:
      for i in range(1, 5):
        self.set_speed((i, 0))

  def close(self) -> None:
    """Closes the I2C connection to the motor hat."""

    if self._bus is not None and self._backend == 'Pi4':
      self.log(logging.INFO, "Closing the connection to the Motor Hat")
      self._bus.close()

  def _set_motor(self, nr: int, cmd: float) -> None:
    """Sets the PWMs associated with a given motor.

    Args:
      nr: The index of the motor to drive.
      cmd: The command as a :obj:`float`, that should be between -1 and 1.
    """

    if self._backend == 'Pi4':

      # Special settings for the extreme values
      if cmd == 0:
        self._write_i2c(motor_hat_pos[nr], motor_hat_0xFF)
        self._write_i2c(motor_hat_neg[nr], motor_hat_0xFF)
      elif cmd == 1.0:
        self._write_i2c(motor_hat_pos[nr], motor_hat_0xFF)
        self._write_i2c(motor_hat_neg[nr], [0x00 for _ in range(4)])
      elif cmd == -1.0:
        self._write_i2c(motor_hat_pos[nr], [0x00 for _ in range(4)])
        self._write_i2c(motor_hat_neg[nr], motor_hat_0xFF)

      # The positive line is driven for positive commands, and reversely
      elif cmd > 0:
        duty_cycle = (int(0xFFFF * cmd) + 1) >> 4
        pack_into("<HH", self._buf, 0, *(0, duty_cycle))
        self._write_i2c(motor_hat_pos[nr], self._buf)
        self._write_i2c(motor_hat_neg[nr], [0x00 for _ in range(4)])
      else:
        duty_cycle = (int(0xFFFF * abs(cmd)) + 1) >> 4
        pack_into("<HH", self._buf, 0, *(0, duty_cycle))
        self._write_i2c(motor_hat_pos[nr], [0x00 for _ in range(4)])
        self._write_i2c(motor_hat_neg[nr], self._buf)

    # Blinka handles all the job internally
    elif self._backend == 'blinka':
      setattr(getattr(self._bus, f'motor{nr}'), 'throttle', cmd)

  def _write_i2c(self, register: int, buf: Union[list, bytearray]) -> None:
    """Thin wrapper to reduce verbosity."""

    self._bus.write_i2c_block_data(self._address, register, list(buf))



# coding: utf-8

import numpy as np
from typing import Optional, List
import logging
import logging.handlers

from .camera_process import CameraProcess
from ...tool.camera_config import Box, SpotsBoxes
from ...tool.image_processing import DISCorrelTool


class DISCorrelProcess(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can perform
  Dense Inverse Search on a given ROI in the acquired images, and calculate
  various fields from it.

  It is used by the :class:`~crappy.blocks.DISCorrel` Block to parallelize the
  image processing and the image acquisition. It delegates most of the
  computation to the :obj:`~crappy.tool.image_processing.DISCorrelTool`. It is
  from this class that the output values are sent to the downstream Blocks, and
  that the :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` are sent
  to the :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess for
  display.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               patch: Box,
               fields: Optional[List[str]] = None,
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               init: bool = True,
               patch_size: int = 8,
               patch_stride: int = 3,
               residual: bool = False) -> None:
    """Sets the arguments and initializes the parent class.
    
    Args:
      patch: An instance of the
        :class:`~crappy.tool.camera_config.config_tools.Box` class, containing
        the coordinates of the ROI to perform DIS on. This argument is passed
        to the :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used
        in this class.
      fields: The base of fields to use for the projection, given as a
        :obj:`list` of :obj:`str`. The available fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

        This argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      alpha: Weight of the smoothness term in DISFlow, as a :obj:`float`. This
        argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      delta: Weight of the color constancy term in DISFlow, as a :obj:`float`.
        This argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      gamma: Weight of the gradient constancy term in DISFlow , as a
        :obj:`float`. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DISFlow (`0` means full scale), as an :obj:`int`. This
        argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      init: If :obj:`True`, the last field is used to initialize the
        calculation for the next one. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DISFlow, as an :obj:`int`. This argument
        is passed to the :obj:`~crappy.tool.image_processing.DISCorrelTool` and
        not used in this class.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DISFlow, as an :obj:`int`. This
        argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      patch_size: Size of an image patch for matching in DISFlow
        (in pixels). This argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      patch_stride: Stride between neighbor patches in DISFlow. Must be
        less than patch size. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DISCorrelTool` and not used in this
        class.
      residual: If :obj:`True`, the residuals will be computed at each new
        frame and sent to downstream Blocks, by default under the ``'res'``
        label.
    """

    super().__init__()

    # Arguments to pass to the DISCorrelTool
    self._box = patch
    self._fields = fields
    self._alpha = alpha
    self._delta = delta
    self._gamma = gamma
    self._finest_scale = finest_scale
    self._init = init
    self._iterations = iterations
    self._gradient_iterations = gradient_iterations
    self._patch_size = patch_size
    self._patch_stride = patch_stride
    
    # Other attributes
    self._residual = residual
    self._dis_correl: Optional[DISCorrelTool] = None
    self._img0_set = False

  def init(self) -> None:
    """Instantiates the :obj:`~crappy.tool.image_processing.DISCorrelTool` that
    will perform the Dense Inverse Search."""

    self.log(logging.INFO, "Instantiating the Discorrel tool")
    self._dis_correl = DISCorrelTool(
        box=self._box,
        fields=self._fields,
        alpha=self._alpha,
        delta=self._delta,
        gamma=self._gamma,
        finest_scale=self._finest_scale,
        init=self._init,
        iterations=self._iterations,
        gradient_iterations=self._gradient_iterations,
        patch_size=self._patch_size,
        patch_stride=self._patch_stride)
    self._dis_correl.set_box()

  def loop(self) -> None:
    """This method grabs the latest frame and gives it for processing to the
    :obj:`~crappy.tool.image_processing.DISCorrelTool`. Then sends the result
    of the dense inverse search to the downstream Blocks.

    If there's no new frame grabbed, doesn't do anything. On the first acquired
    frame, does not process it but initializes the DISCorrelTool with it
    instead. Also sends the ROI for display to the
    :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess.
    """

    # On the first frame, initializes the dense inverse search
    if not self._img0_set:
      self.log(logging.INFO, "Setting the reference image")
      self._dis_correl.set_img0(np.copy(self.img))
      self._img0_set = True
      return

    # Calculating the fields and sending them to downstream Blocks
    self.log(logging.DEBUG, "Processing the received image")
    data = self._dis_correl.get_data(self.img, self._residual)
    self.send([self.metadata['t(s)'], self.metadata, *data])

    # Sending the ROI to the Displayer for display
    self.send_to_draw(SpotsBoxes(self._dis_correl.box))



# coding: utf-8

import numpy as np
from typing import Dict, Any, Union, Iterable
import logging

from .meta_modifier import Modifier


class Demux(Modifier):
  """Modifier converting a stream into a regular data flow interpretable by
  most Blocks.

  It is meant to be used on a :class:`~crappy.links.Link` taking an
  :class:`~crappy.blocks.IOBlock` in streamer mode as an input. It converts
  the stream to make it readable by most Blocks, and also splits the stream in
  several labels if necessary.

  It takes a stream as an input, i.e. a :obj:`dict` whose values are
  :obj:`numpy.array`, and outputs another :obj:`dict` whose values are
  :obj:`float`. If the numpy arrays contains several columns (corresponding to
  several acquired channels), it splits them into several labels.

  Important:
    In the process of converting the stream data to regular labeled data, much
    information is lost ! This Modifier is intended to format the stream data
    for low-frequency plotting, or low-frequency decision-making. To save all
    the stream data, use the :class:`~crappy.blocks.HDFRecorder` Block.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               labels: Union[str, Iterable[str]],
               stream_label: str = "stream",
               mean: bool = False,
               time_label: str = "t(s)",
               transpose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      labels: The labels corresponding to the rows or columns of the stream.
        It can be either a single label, or an iterable of labels (like a
        :obj:`list` or a :obj:`tuple`). They must be given in the same order as
        they appear in the stream. If fewer labels are given than there are
        rows or columns in the stream, only the data from the first rows or
        columns will be retrieved.
      stream_label: The label carrying the stream.
      mean: If :obj:`True`, the returned value will be the average of the
        row or column. Otherwise, it will be the first value.
      time_label: The label carrying the time information.
      transpose: If :obj:`True`, each label corresponds to a row in the stream.
        Otherwise, a label corresponds to a column in the stream.
    
    .. versionchanged:: 1.5.10 renamed *stream* argument to *stream_label*
    """

    super().__init__()

    if isinstance(labels, str):
      labels = (labels,)
    self._labels = labels

    self._stream_label = stream_label
    self._mean = mean
    self._time_label = time_label
    self._transpose = transpose

  def __call__(self, data: Dict[str, np.ndarray]) -> Dict[str, Any]:
    """Retrieves for each label its value in the stream, also gets the
    corresponding timestamp, and returns them.
    
    .. versionchanged:: 1.5.10 
       merge evaluate_mean and evaluate_nomean methods into evaluate
    .. versionchanged:: 2.0.0 renamed from evaluate to __call__
    """

    self.log(logging.DEBUG, f"Received {data}")

    # If there are no rows or no column, cannot perform the demux
    if 0 in data[self._stream_label].shape:
      return data

    # Getting either the average or the first value for each label
    for i, label in enumerate(self._labels):
      # The data of a given label is on a same row
      if self._transpose:
        if self._mean:
          data[label] = np.mean(data[self._stream_label][i, :])
        else:
          data[label] = data[self._stream_label][i, 0]
      # The data of a given label is on a same column
      else:
        if self._mean:
          data[label] = np.mean(data[self._stream_label][:, i])
        else:
          data[label] = data[self._stream_label][0, i]

    # Discarding the raw data
    del data[self._stream_label]

    # Keeping either the average or the first time value
    if self._mean:
      data[self._time_label] = np.mean(data[self._time_label])
    else:
      data[self._time_label] = np.squeeze(data[self._time_label])[0]

    self.log(logging.DEBUG, f"Sending {data}")

    return data



# coding: utf-8

from multiprocessing import Process, get_start_method
from multiprocessing.connection import Connection
from multiprocessing.queues import Queue
import numpy as np
from typing import Optional, Union
from time import time
from select import select
import logging
import logging.handlers
from platform import system

from ...camera_config import Box
from ...._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")
try:
  from skimage.filters import threshold_otsu
except (ModuleNotFoundError, ImportError):
  threshold_otsu = OptionalModule("skimage", "Please install scikit-image to "
                                             "use Video-extenso")


class LostSpotError(Exception):
  """Exception raised when a spot is lost, or when there's too much
  overlapping."""


class Tracker(Process):
  """:obj:`multiprocessing.Process` whose task is to track a spot on an image.

  It receives a subframe centered on the last known position of the spot, and
  returns the updated position of the detected spot. It is meant to be used in
  association with the 
  :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`.
  
  .. versionadded:: 2.0.0
  """

  names = list()

  def __init__(self,
               pipe: Connection,
               logger_name: str,
               log_level: Optional[int],
               log_queue: Queue,
               white_spots: bool = False,
               thresh: Optional[int] = None,
               blur: Optional[float] = 5) -> None:
    """Sets the arguments.

    Args:
      pipe: The :obj:`~multiprocessing.connection.Connection` object through
        which the image is received and the updated coordinates of the spot
        are sent back.
      logger_name: The name of the parent :obj:`~logging.Logger` as a 
        :obj:`str`, used for naming the Logger in this class.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.
      white_spots: If :obj:`True`, detects white objects on a black background,
        else black objects on a white background.
      thresh: If given, this threshold value will always be used for isolating
        the spot from the background. If not given (:obj:`None`), a new
        threshold is recalculated for each new subframe. Spots are less likely
        to be lost with an adaptive threshold, but it takes a bit more time.
      blur: If not :obj:`None`, the subframe is first blurred before trying to
        detect the spot. This argument gives the size of the kernel to use for
        blurring. Better results are obtained with blurring, but it takes a bit
        more time.
    """

    super().__init__()
    self.name = self.get_name(logger_name, type(self).__name__)
    self._system = system()

    self._pipe = pipe
    self._white_spots = white_spots
    self._thresh = thresh
    self._blur = blur

    self._logger: Optional[logging.Logger] = None
    self._log_level = log_level
    self._log_queue = log_queue

    self._n = 0
    self._last_warn = time()

  @classmethod
  def get_name(cls, logger_name: str, self_name: str) -> str:
    """Method for naming the Tracker processes so that each of them has a
    unique name.

    Args:
      logger_name: The name of the parent :obj:`~logging.Logger`, as a 
        :obj:`str`.
      self_name: The name of the current class.

    Returns:
      The chosen name for the current Tracker process, as a :obj:`str`.
    """

    i = 1
    while f"{logger_name}.{self_name}-{i}" in cls.names:
      i += 1

    cls.names.append(f"{logger_name}.{self_name}-{i}")
    return f"{logger_name}.{self_name}-{i}"

  def run(self) -> None:
    """Continuously reads incoming subframes, tries to detect a spot and sends
    back the coordinates of the detected spot.

    Can only be stopped either with a :exc:`KeyboardInterrupt` or when
    receiving a text message from the 
    :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`.
    """

    # Looping forever for receiving data
    try:
      self._set_logger()

      while True:
        # Making sure the call to recv is not blocking
        if self._pipe.poll(0.5):
          y_start, x_start, img = self._pipe.recv()
          self._log(logging.DEBUG, "Received data from pipe")
          self._n += 1

          # If a string is received, always means the process has to stop
          if isinstance(img, str):
            break

          # Simply sending back the new Box containing the spot
          try:
            self._log(logging.DEBUG, "Sending back data through pipe")
            self._send(self._evaluate(x_start, y_start, img))

          # If the caught exception is a KeyboardInterrupt, simply stopping
          except KeyboardInterrupt:
            self._log(logging.INFO, "Caught KeyboardInterrupt, stopping the "
                                    "process")
            break
          # Sending back the exception if anything else unexpected happened
          except (Exception,) as exc:
            self._logger.exception("Caught exception while tracking spot",
                                   exc_info=exc)
            self._send('stop')
            break

    # In case the user presses CTRL+C, simply stopping the process
    except KeyboardInterrupt:
      self._log(logging.INFO, "Caught KeyboardInterrupt, stopping the process")

  def _evaluate(self, x_start: int, y_start: int, img: np.ndarray) -> Box:
    """Takes a sub-image, applies a threshold on it and tries to detect the new
    position of the spot.

    Args:
      x_start: The x position of the top left pixel of the subframe on the
        entire image.
      y_start: The y position of the top left pixel of the subframe on the
        entire image.
      img: The subframe on which to search for a spot.

    Returns:
      A :class:`~crappy.tool.camera_config.config_tools.Box` object containing 
      the x and y start and end positions of the detected spot, as well as the 
      coordinates of the centroid.
    """

    # First, blurring the image if asked to
    if self._blur is not None and self._blur > 1:
      img = cv2.medianBlur(img, self._blur)

    # Determining the best threshold for the image if required
    thresh = self._thresh if self._thresh is not None else threshold_otsu(img)

    # Getting all pixels superior or inferior to threshold
    if self._white_spots:
      black_white = (img > thresh).astype('uint8')
    else:
      black_white = (img <= thresh).astype('uint8')

    # Checking that the detected spot is large enough
    if np.count_nonzero(black_white) < 0.1 * img.size:

      # If the threshold is pre-defined, trying again with an updated one
      if self._thresh is not None:
        self._log(logging.WARNING,
                  "Detected spot too small compared with overall box size, "
                  "recalculating threshold")
        thresh = threshold_otsu(img)
        if self._white_spots:
          black_white = (img > thresh).astype('uint8')
        else:
          black_white = (img <= thresh).astype('uint8')

        # If the spot still cannot be detected, aborting
        if np.count_nonzero(black_white) < 0.1 * img.size:
          self._log(logging.ERROR,
                    "Couldn't detect spot with adaptive threshold, aborting !")
          raise LostSpotError

      # If an adaptive threshold is already used, nothing more can be done
      else:
        self._log(logging.ERROR,
                  "Couldn't detect spot with adaptive threshold, aborting !")
        raise LostSpotError

    # Calculating the coordinates of the centroid using the image moments
    moments = cv2.moments(black_white)
    try:
      x = moments['m10'] / moments['m00']
      y = moments['m01'] / moments['m00']
    except ZeroDivisionError:
      raise ZeroDivisionError("Couldn't compute the centroid because the "
                              "moment of order 0, 0 is zero !")

    # Getting the updated centroid and coordinates of the spot
    x_min, y_min, width, height = cv2.boundingRect(black_white)
    return Box(x_start=x_start + x_min,
               y_start=y_start + y_min,
               x_end=x_start + x_min + width,
               y_end=y_start + y_min + height,
               x_centroid=x_start + x,
               y_centroid=y_start + y)

  def _send(self, val: Union[Box, str]) -> None:
    """Sends a message to the 
    :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`, and 
    in Linux checks that the :obj:`multiprocessing.Pipe` is not full before 
    sending."""

    if self._system == 'Linux':
      if select([], [self._pipe], [], 0)[1]:
        self._pipe.send(val)
      elif time() - self._last_warn > 1:
          self._last_warn = time()
          self._log(logging.WARNING, f"Cannot send the detected spot to the "
                                     f"VideoExtenso tool, the Pipe is full !")
    else:
      self._pipe.send(val)

  def _set_logger(self) -> None:
    """Instantiates and sets up the logger for the instance."""

    logger = logging.getLogger(self.name)

    # Disabling logging if requested
    if self._log_level is not None:
      logger.setLevel(self._log_level)
    else:
      logging.disable()

    # On Windows, the messages need to be sent through a Queue for logging
    if get_start_method() == "spawn" and self._log_level is not None:
      queue_handler = logging.handlers.QueueHandler(self._log_queue)
      queue_handler.setLevel(self._log_level)
      logger.addHandler(queue_handler)

    self._logger = logger

  def _log(self, level: int, msg: str) -> None:
    """Sends a log message to the :obj:`~logging.Logger`.

    Args:
      level: The logging level, as an :obj:`int`.
      msg: The message to log, as a :obj:`str`.
    """

    if self._logger is None:
      return
    self._logger.log(level, msg)



# coding: utf-8

from multiprocessing import Process, managers, get_start_method, \
  current_process
from multiprocessing.synchronize import Event, RLock, Barrier
from multiprocessing.sharedctypes import SynchronizedArray
from multiprocessing.connection import Connection
from multiprocessing.queues import Queue
from threading import BrokenBarrierError
import numpy as np
from typing import Optional, Tuple, List, Union, Dict, Any, Iterable
import logging
import logging.handlers
from select import select
from time import time
from platform import system

from ...links import Link
from ..._global import LinkDataError
from ...tool.camera_config import Overlay


class CameraProcess(Process):
  """This :obj:`~multiprocessing.Process` is the base class for all the helper
  Processes of the :class:`~crappy.blocks.Camera` Block.
  
  It defines a base architecture that all the children classes use, and on top
  of which they add the specific actions they perform. This class is not meant 
  to be instantiated as is, it should always be subclassed.
  
  This class and its children can be seen as a tool for the 
  :class:`~crappy.blocks.Camera` Block and its children. Several instances can
  be started by the Camera Block, each instance performing a different action
  with the acquired images. This allows to parallelize image acquisition, 
  display, processing and recording on multiple CPU cores, to increase the 
  achieved FPS.
  
  The Camera Block performs the acquisition, and makes the latest captured
  image available to all the CameraProcess :obj:`~multiprocessing.Process` 
  through an :obj:`~multiprocessing.Array`. They are then free to run at their
  own rhythm, but are sure to always grab the latest available frame.
  
  The instantiation, startup and termination of the CameraProcesses is all
  managed by the parent :class:`~crappy.blocks.Camera` Block, depending on the
  provided arguments. Users should normally not need to call this class
  themselves.
  
  .. versionadded:: 2.0.0
  """

  def __init__(self) -> None:
    """Initializes the parent class and all the instance attributes."""

    super().__init__()
    self.name = f"{current_process().name}.{type(self).__name__}"
    self._system = system()

    # Logging-related objects
    self._log_queue: Optional[Queue] = None
    self._logger: Optional[logging.Logger] = None
    self._log_level: Optional[int] = None

    # These objects will be shared later by the Camera Block
    self._img_array: Optional[SynchronizedArray] = None
    self._data_dict: Optional[managers.DictProxy] = None
    self._lock: Optional[RLock] = None
    self._cam_barrier: Optional[Barrier] = None
    self._stop_event: Optional[Event] = None
    self._shape: Optional[Tuple[int, int]] = None
    self._to_draw_conn: Optional[Connection] = None
    self._outputs: List[Link] = list()
    self._labels: List[str] = list()
    self.img: Optional[np.ndarray] = None
    self._dtype = None
    self.metadata = {'ImageUniqueID': None}
    self._img0_set = False

    # Other attribute for internal use
    self._last_warn = time()
    self.fps_count = 0
    self._display_freq: Optional[bool] = None
    self._last_fps = time()

  def set_shared(self,
                 array: SynchronizedArray,
                 data_dict: managers.DictProxy,
                 lock: RLock,
                 barrier: Barrier,
                 event: Event,
                 shape: Tuple[int, int],
                 dtype,
                 to_draw_conn: Optional[Connection],
                 outputs: List[Link],
                 labels: Optional[List[str]],
                 log_queue: Queue,
                 log_level: Optional[int] = 20,
                 display_freq: bool = False) -> None:
    """Method allowing the :class:`~crappy.blocks.Camera` Block to share
    :mod:`multiprocessing` synchronization objects with this class.
    
    Args:
      array: The :obj:`~multiprocessing.Array` containing the last frame
        acquired by the Camera Block.
      data_dict: A :obj:`dict` managed by a :obj:`~multiprocessing.Manager` and
        containing the metadata of the last acquired frame.
      lock: A :obj:`~multiprocessing.RLock` ensuring that the CameraProcess and
        the Camera Block do not try to access the shared array at the same
        time.
      barrier: A :obj:`~multiprocessing.Barrier` ensuring that all the
        CameraProcesses wait for a start signal from the Camera Block before
        starting to run.
      event: A :obj:`~multiprocessing.Event` indicating to the CameraProcess
        when to stop running. It is either set by the Camera Block, or by a
        CameraProcess.
      shape: The expected shape of the image, as a :obj:`tuple`. It is 
        necessary as the frames are shared as a one-dimensional array.
      dtype: The expected dtype of the image. It is necessary for 
        reconstructing the image from the one-dimensional shared array.
      to_draw_conn: A :obj:`~multiprocessing.Connection` for sending or
        receiving :class:`~crappy.tool.camera_config.config_tools.Overlay`
        objects to draw on top of the displayed image.
      outputs: The :class:`~crappy.links.Link` objects for sending data to
        downstream Blocks. They are the same as those owned by the Camera 
        Block.
      labels: The labels to use when sending data to downstream Blocks.
      log_queue: A :obj:`~multiprocessing.Queue` for sending the log messages
        to the main :obj:`~logging.Logger`, only used in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      display_freq: If :obj:`True`, the looping frequency of this class will be
        displayed while running.
    """

    self._img_array = array
    self._data_dict = data_dict
    self._lock = lock
    self._cam_barrier = barrier
    self._stop_event = event
    self._shape = shape
    self._dtype = dtype
    self._to_draw_conn = to_draw_conn
    self._outputs = outputs
    self._labels = labels

    # Logging related attributes
    self._log_queue = log_queue
    self._log_level = log_level
    self._display_freq = display_freq

    self.img = np.empty(shape=shape, dtype=dtype)

  def run(self) -> None:
    """This method is the core of the :obj:`~multiprocessing.Process`.
    
    It starts by initializing the :obj:`~logging.Logger`, and then performs any
    additional action required before processing images. Once all the
    CameraProcesses are ready, it loops forever and processes images until told
    to stop or an exception is raised. And finally, it performs any action
    required for properly exiting.

    This method is quite similar to the :meth:`~crappy.blocks.Block.run` method
    of the :class:`~crappy.blocks.Block`, although it is much simpler.
    """

    try:
      # First thing, setting the Logger
      self._set_logger()
      self.log(logging.INFO, "Logger configured")

      # Initializing the CameraProcess, and breaking the Barrier to warn the
      # other CameraProcesses in case something goes wrong
      try:
        self.init()
      except (Exception,):
        self._cam_barrier.abort()
        self.log(logging.ERROR, "Breaking the barrier due to caught exception"
                                " while preparing")
        raise

      # Waiting for all other CameraProcess to be ready
      self.log(logging.INFO, "Waiting for the other Camera processes to be "
                             "ready")
      self._cam_barrier.wait()
      self.log(logging.INFO, "All Camera processes ready now")

      self._last_fps = time()

      # Looping forever until told to stop or an exception is raised
      while not self._stop_event.is_set():
        # Only looping if a new image is available
        if self._get_data():
          self.log(logging.DEBUG, "Running the loop method")
          self.loop()
          self.fps_count += 1

        # Displaying the looping frequency is required
        if self._display_freq:
          t = time()
          if t - self._last_fps > 2:
            self.log(logging.INFO, f"Images processed /s: "
                                   f"{self.fps_count / (t - self._last_fps)}")
            self._last_fps = t
            self.fps_count = 0

      self.log(logging.INFO, "Stop event set, stopping the processing")

    # Case when CTRL+C was pressed
    except KeyboardInterrupt:
      self.log(logging.INFO, "KeyboardInterrupt caught, stopping the "
                             "processing")

    # Case when another CameraProcess raised an exception while initializing
    except BrokenBarrierError:
      self.log(logging.WARNING,
               "Exception raised in another Camera process while waiting "
               "for all Camera processes to be ready, stopping")

    # Handling any other unexpected exception
    except (Exception,) as exc:
      self._logger.exception("Exception caught wile running !", exc_info=exc)
      self.log(logging.ERROR, "Setting the stop event to stop the other "
                              "Camera processes")
      self._stop_event.set()
      raise

    # Always calling finish in the end
    finally:
      self.finish()

  def init(self) -> None:
    """This method should perform any action required for initializing the
    CameraProcess.

    It is meant to be overwritten by children classes, at is otherwise does not
    perform any action.
    
    It is called right after the :obj:`~multiprocessing.Process` starts, and
    when the images haven't started to be acquired yet.
    """

    ...

  def loop(self) -> None:
    """This method is the main loop of the CameraProcess.
    
    It is called repeatedly until the :obj:`~multiprocessing.Process` is told
    to stop. It should perform the desired action for handling the latest
    available frame, stored in the *self.img* attribute. The latest available
    metadata containing at least the timestamp and frame index of the latest
    image is stored in *self.metadata*.
    
    This method is meant to be overwritten by children classes, at is otherwise 
    does not perform any action.
    """

    ...

  def finish(self) -> None:
    """This method should perform any action required for properly exiting the
    CameraProcess.

    It is meant to be overwritten by children classes, at is otherwise does not
    perform any action.

    It is the last method called before the :obj:`~multiprocessing.Process`
    ends, and at that point no more images are being acquired.
    """

    ...

  def send(self, data: Optional[Union[Dict[str, Any],
                                      Iterable[Any]]]) -> None:
    """This method allows sending data to downstream Blocks.

    It is similar to the :meth:`~crappy.blocks.Block.send` method of the
    :class:`~crappy.blocks.Block`. It accepts data either as a :obj:`dict`, or
    as an iterable of values.
    """

    # Just in case, not handling non-existing data
    if data is None:
      return

    # Case when the data to send is not given as a dict
    if not isinstance(data, dict):
      # First, checking that labels are provided
      if self._labels is None or not self._labels:
        self._logger.log(logging.ERROR, "Trying to send data as an iterable, "
                                        "but no labels are specified !")
        raise LinkDataError

      # Trying to convert iterable data to dict using the given labels
      try:
        self._logger.log(logging.DEBUG, f"Converting {data} to dict before "
                                        f"sending")
        data = dict(zip(self._labels, data))
      except TypeError:
        self._logger.log(logging.ERROR, f"Cannot convert data to send (of type"
                                        f" {type(data)}) to dict ! Please "
                                        f"ensure that the data is given as an "
                                        f"iterable, as well as the labels.")

    # Sending the data to the downstream Blocks
    for link in self._outputs:
      self._logger.log(logging.DEBUG, f"Sending {data} to Link {link.name}")
      link.send(data)

  def send_to_draw(self, to_draw: Iterable[Overlay]) -> None:
    """This method sends a collection of
    :class:`~crappy.tool.camera_config.config_tools.Overlay` objects to the
    :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess.

    The overlays are sent by the CameraProcess performing the image processing,
    so that the area(s) of interest can be displayed simultaneously.
    """

    # Not sending if there's no Connection to send data through
    if self._to_draw_conn is None:
      return

    self.log(logging.DEBUG, "Sending the overlays to the displayer process")

    # Sending the overlay
    if self._system == 'Linux':
      if select([], [self._to_draw_conn], [], 0)[1]:
        # Can only check on Linux if a pipe is full
        self._to_draw_conn.send(to_draw)
      elif time() - self._last_warn > 1:
        # Warning in case the pipe is full
        self._last_warn = time()
        self.log(logging.WARNING, f"Cannot send the overlay to draw to the "
                                  f"Displayer process, the Pipe is full !")
    else:
      self._to_draw_conn.send(to_draw)

  def log(self, level: int, msg: str) -> None:
    """Sends a log message to the :obj:`~logging.Logger`.

    Args:
      level: The logging level, as an :obj:`int`.
      msg: The message to log, as a :obj:`str`.
    """

    if self._logger is None:
      return
    self._logger.log(level, msg)

  def _get_data(self) -> bool:
    """This method allows to grab the latest available frame.

    It first acquired the :obj:`~multiprocessing.RLock` protecting the shared
    :obj:`~multiprocessing.Array` containing the image, then copies the image
    locally and releases the Lock. It also copies the metadata associated to
    the image.

    Returns:
      :obj:`True` in case a frame was acquired and needs to be handled, or
      :obj:`False` if no frame was grabbed and nothing should be done.
    """

    # Acquiring the Lock to avoid conflicts with other CameraProcesses
    with self._lock:

      # In case there's no frame grabbed yet
      if 'ImageUniqueID' not in self._data_dict:
        return False

      # In case the frame in buffer was already handled during a previous loop
      if self._data_dict['ImageUniqueID'] == self.metadata['ImageUniqueID']:
        return False

      # Copying the metadata
      self.metadata = self._data_dict.copy()

      self.log(logging.DEBUG, f"Got new image to process with id "
                              f"{self.metadata['ImageUniqueID']}")

      # Copying the frame
      np.copyto(self.img,
                np.frombuffer(self._img_array.get_obj(),
                              dtype=self._dtype).reshape(self._shape))

    return True

  def _set_logger(self) -> None:
    """Initializes the :obj:`~logging.Logger` for the CameraProcess.

    If :func:`multiprocessing.get_start_method` is `'spawn'` (mostly Windows
    and Mac for Python < 3.12), redirects the log messages to a
    :obj:`multiprocessing.Queue` sending them to the main Process.
    """

    logger = logging.getLogger(self.name)

    # Disabling logging if requested
    if self._log_level is not None:
      logger.setLevel(self._log_level)
    else:
      logging.disable()

    # On Windows, the messages need to be sent through a Queue for logging
    if get_start_method() == "spawn" and self._log_level is not None:
      queue_handler = logging.handlers.QueueHandler(self._log_queue)
      queue_handler.setLevel(min(self._log_level, logging.INFO))
      logger.addHandler(queue_handler)

    self._logger = logger



# coding: utf-8

from .camera_process import CameraProcess
from .dis_correl import DISCorrelProcess
from .display import Displayer
from .dic_ve import DICVEProcess
from .gpu_correl import GPUCorrelProcess
from .gpu_ve import GPUVEProcess
from .record import ImageSaver
from .video_extenso import VideoExtensoProcess
